"""Laser Beams
================

Here we implement the generic ``LaserBeam`` class, as well as some actual
laser beam classes

Examples
---------

Setup a Gaussian beam

.. code-block:: python

    from atomsmltr.environment.lasers import GaussianLaserBeam
    from atomsmltr.environment.lasers.polarization import CircularLeft


    beam = GaussianLaserBeam(
        wavelength=399e-9,
        waist=50e-6,
        power=30e-3,
        waist_position=(0, 0, 0),
        direction=(0, 0, 1),
        polarization=CircularLeft(),
    )

See also
--------
atomsmltr.environment.lasers.polarization

"""

# % IMPORTS
import numpy as np
import matplotlib.pyplot as plt

from abc import abstractmethod

# % LOCAL IMPORTS
from .polarization import Vertical, Polarization
from ..envbase import EnvObject
from ...utils.infostring import InfoString


# % GLOBAL DEFINITIONS

DIRECTION_TYPES = ["vector", "thetaphi"]  # allowed values for `direction_type``


# % TOOL FUNCTIONS


def _intensity_gauss(
    r: float, z: float, w0: float, P: float, wavelength: float
) -> float:
    """Computes intensity for a Gaussian beam of waist w0 and power P0 at position
    (r, z), in cynlindrical coordinates. The beam is propagating along z, and the waist
    is located at r = z = 0. Lengths should be given in meters, and powers in watts.
    Intensity is returned in W/m^2

    Args:
        r (float): radial coordinate (distance to beam axis) in _meters_
        z (float): axial coordinate (distance to beam waist) in _meters_
        w0 (float): Gaussian beam waist radius (1/e^2) in _meters_
        P (float): laser power in _Watts_
        wavelength (float): laser wavelength in _meters_

    Returns:
        intensity (float): laser intensity in _W/m^2_
    """

    zR = np.pi * w0**2 / wavelength
    wz = w0 * np.sqrt(1 + z**2 / zR**2)
    I0 = 2 * P / np.pi / w0**2
    intensity = I0 * (w0 / wz) ** 2 * np.exp(-2 * (r**2) / wz**2)
    return intensity


# % ABSTRACT CLASSES


class LaserBeam(EnvObject):
    """Representing laser beams

    Parameters
    ----------
    wavelength : float, optional
        vacuum wavelength (m), by default 399e-9
    waist : float, optional
        1/e^2 waist radius (m), by default 1e-3
    power : float, optional
        laser power (W), by default 1e-3
    waist_position : array, shape (,3), optional
        cartesian coordinates of the waist / focus position,
        in meters and in the lab frame, by default (0, 0, 0)
    direction : array, shape (,3) or (,2), optional
        depending on 'direction type', a vector or a (theta, phi) couple
        giving the propagation direction of the beam
    direction_type : str, optional
        type of direction : "vector" or "thetaphi", by default "vector"
    polarization : Polarization, optional
        laser polarization, by default Vertical()
    tag : str, optional
        laser tag, by default None
    """

    def __init__(
        self,
        wavelength: float = 399e-9,
        waist: float = 1e-3,
        power: float = 1e-3,
        waist_position: np.ndarray = (0, 0, 0),
        direction: np.ndarray = (0, 0, 1),
        direction_type: str = "vector",
        polarization: Polarization = Vertical(),
        tag: str = None,
    ):

        self.wavelength = wavelength
        self.waist = waist
        self.power = power
        self.waist_position = waist_position
        # /!\ direction_type has to be defined BEFORE direction !!
        self.direction_type = direction_type
        self.direction = direction
        self.polarization = polarization

        super(LaserBeam, self).__init__(tag=tag)

    # -- REQUESTED PROPERTY FOR ENVOBJECTS
    @property
    def vector(self):
        return False

    # -- COMMON METHODS DEFINED HERE
    def _convert_coordinates_to_laser_frame(
        self, position: np.ndarray, nocheck=False
    ) -> np.ndarray:
        """Converts lab frame cartesian coordinates to laser frame coordinates.

        Parameters
        ----------
        position : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates in the lab frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        position_laser : array of shape (3,) or (n1, n2, ..., 3) (same as position)
            cartesian coordinates in the laser frame

        Notes
        -------

        'position' should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        The laser frame is centered at the laser waist, and has the z axis aligned
        with the laser propagation.

        The unit vector defining laser propagation is defined with two angles, theta
        and phi : theta is the angle between the unit vector and the z axis of the lab
        frame, and phi is the angle of the unit vector project on the (x, y) plane of the
        lab frame, w.r.t the x axis.

        To define the new coordinates (x_laser, y_laser, z_laser) in the laser frame, we
        proceed as follow:

        1) we shift the frame to center it on the waist position:
            (x, y, z) > (xc, yc, zc)
        2) we perform a rotation with an angle phi around the lab frame z axis:
            (xc, yc, zc) > (x', y', z')
        3) we perform a rotation with an angle theta around the y' axis of the new frame:
            (x', y', z') > (x_laser, y_laser, z_laser)

        For convenience reasons, we also return polar coordinates in the laser frame


        Note
        -----

            Note: in some cases (elliptical beams for instance) it might be interesting to include
            a final rotation around the laser propagation axis in the laser frame. We decided that
            this rotation will be handled in the `intensity()` method of the corresponding class.
        """

        # convert to array if needed
        position = self._check_position_array(position, nocheck)
        # get coordinates
        x, y, z = position.T
        # shift center
        x0, y0, z0 = self._waist_position
        xc = x - x0
        yc = y - y0
        zc = z - z0

        # rotate : phi around z axis, then theta along new y axis
        # see function docstring and documentation for rotation & frames definitions
        theta = self._unit_vector_theta
        phi = self._unit_vector_phi
        x_laser = (
            xc * np.cos(theta) * np.cos(phi)
            + yc * np.cos(theta) * np.sin(phi)
            - zc * np.sin(theta)
        )
        y_laser = -xc * np.sin(phi) + yc * np.cos(phi)
        z_laser = (
            xc * np.sin(theta) * np.cos(phi)
            + yc * np.sin(theta) * np.sin(phi)
            + zc * np.cos(theta)
        )

        # also yield cylindrical coordinates - NOT ANYMORE
        # rho_laser = np.sqrt(x_laser**2 + y_laser**2)
        # th_laser = np.arctan2(y_laser, x_laser)
        position_laser = np.array([x_laser, y_laser, z_laser]).T
        return position_laser

    def _convert_vector_to_laser_frame(
        self, vec: np.ndarray, nocheck: bool = False
    ) -> np.ndarray:
        """Rotates a vector from lab frame to laser frame.

        Parameters
        ----------
        vec : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the vectors in the lab frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        vec_laser : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the vectors in the laser frame

        Notes
        -------

        'vec' should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains vector coordinates x, y, z

        The unit vector defining laser propagation is defined with two angles, theta
        and phi : theta is the angle between the unit vector and the z axis of the lab
        frame, and phi is the angle of the unit vector project on the (x, y) plane of the
        lab frame, w.r.t the x axis.

        To perform a rotation from lab frame (x, y, z) to laser frame (x_laser, y_laser, z_laser):

        1) we perform a rotation with an angle phi around the lab frame z axis:
            (x, y, z) > (x', y', z')
        2) we perform a rotation with an angle theta around the y' axis of the new frame:
            (x', y', z') > (x_laser, y_laser, z_laser)

        """

        # convert vec
        vec = self._check_position_array(vec, nocheck)
        x, y, z = vec.T
        # rotate : phi around z axis, then theta along new y axis
        # see function docstring and documentation for rotation & frames definitions
        # shorthands
        costheta = self.__costheta
        sintheta = self.__sintheta
        cosphi = self.__cosphi
        sinphi = self.__sinphi
        # compute
        x_laser = x * costheta * cosphi + y * costheta * sinphi - z * sintheta
        y_laser = -x * sinphi + y * cosphi
        z_laser = x * sintheta * cosphi + y * sintheta * sinphi + z * costheta

        vec_laser = np.array([x_laser, y_laser, z_laser]).T
        return vec_laser

    def _convert_vector_to_lab_frame(
        self, vec: np.ndarray, nocheck=False
    ) -> np.ndarray:
        """Rotates a vector from laser frame to lab frame.

        Parameters
        ----------
        vec : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the vectors in the laser frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        vec_lab : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the vectors in the lab frame

        Notes
        ------

        Realizes the reverse operation of `_convert_vector_to_laser_frame`.
        See `_convert_vector_to_laser_frame` docstring for more information

        """

        # convert vec
        vec = self._check_position_array(vec, nocheck)
        x, y, z = vec.T
        # rotate : phi around z axis, then theta along new y axis
        # see function docstring and documentation for rotation & frames definitions
        # shorthands
        costheta = self.__costheta
        sintheta = self.__sintheta
        cosphi = self.__cosphi
        sinphi = self.__sinphi
        # compute
        x_lab = x * costheta * cosphi - y * sinphi + z * sintheta * cosphi
        y_lab = x * costheta * sinphi + y * cosphi + z * sintheta * sinphi
        z_lab = -x * sintheta + z * costheta

        vec_lab = np.array([x_lab, y_lab, z_lab]).T
        return vec_lab

    def get_polarization_vector_in_laser_frame(self) -> np.ndarray:
        """Returns the polarization vector describing the current polarization state, in the **LASER** frame

        Returns
        -------
        p_vec : array of shape (,3)
            cartesian coordinates of the polarization vector (laser frame)

        Notes
        ------
         See documentation for the exact definition of the vector. In short :

        | ``> p_vec = (1, 0, 0)``  : linear polarization along x (vertical)
        | ``> p_vec = (0, 1, 0)``  : linear polarization along y (horizontal)
        | ``> p_vec = (0, 0, 1)``  : circular right polarization
        | ``> p_vec = (0, 0, -1)`` : circular left polarization
        """
        return self.polarization.vector

    def get_polarization_vector_in_lab_frame(self) -> np.ndarray:
        """Returns the polarization vector describing the current polarization state, in the **LAB** frame

        Returns
        -------
        p_vec : array of shape (,3)
            cartesian coordinates of the polarization vector (lab frame)

        Notes
        ------
         See documentation for the exact definition of the vector. In short :

        | ``> p_vec = (1, 0, 0)``  : linear polarization along x (vertical)
        | ``> p_vec = (0, 1, 0)``  : linear polarization along y (horizontal)
        | ``> p_vec = (0, 0, 1)``  : circular right polarization
        | ``> p_vec = (0, 0, -1)`` : circular left polarization
        """
        p_vec_laser_frame = self.polarization.vector
        p_vec_lab_frame = self._convert_vector_to_lab_frame(p_vec_laser_frame)
        return p_vec_lab_frame

    def get_polarization_quant_amplitude(
        self, quantization_axis: np.ndarray, nocheck: bool = False
    ) -> np.ndarray:
        """Returns the projection of the polarization state |Ψ⟩ on |σ+⟩, |σ-⟩ and |π⟩, using
        the vector ``quantization_axis`` as a quantification axis. See documentation
        for a derivation of this projection.

        Parameters
        ----------
        quantization_axis : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the quantization axis vector in the lab frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        polar_amp : array of shape (3,) or (n1, n2, ..., 3)
            contains the polarization amplitude for π, σ+ and σ- components

        Notes
        -----

        The input ``quantization_axis`` should be an array of shape (3,) or (n1, n2, .., 3), where
        the cartesian coordinates of the quantization axis are stored in the last dimension (of size 3)

        the result ``polar_amp`` is an array whose size matches the one of ``quantization_axis``, where the last
        dimension of size 3 contains the projections of the polarization state on π, σ+ and σ-

        That is :

        >>> pi_amp, sigmaplus_amp, sigma_minus_amp = polar_amp.T

        With:

        | ``pi_amp`` =  〈Ψ|π⟩
        | ``sigmaplus_amp`` =  〈Ψ|σ+⟩
        | ``sigma_minus_amp`` =  〈Ψ|σ-⟩

        See Also
        --------
        get_polarization_quant()
        get_polarization_quant_amplitude_dict()
        get_polarization_quant_dict()

        """

        # -- process input
        # - check
        quantization_axis = self._check_position_array(quantization_axis, nocheck)

        # -- compute angles of B field w.r.t k vector, in the laser frame
        # 1) coordinates of uB in laser frame
        uB_laser = self._convert_vector_to_laser_frame(quantization_axis, nocheck)
        # 2) compute angles
        xl, yl, zl = uB_laser.T
        alpha = np.arctan2(np.sqrt(xl * xl + yl * yl), zl)  # polar angle
        beta = np.arctan2(yl, xl)  # azimuthal angle

        # -- get angles of polarization vector in the laser frame
        u, v = self.polarization.get_polarization_vector_angles()

        # -- projections of polarization state |Ψ⟩ on |x⟩ and |y⟩
        # >>> see documentation for explanation
        x_proj = (1 / np.sqrt(2)) * (
            np.exp(-1j * v) * np.cos(u / 2) + np.exp(1j * v) * np.sin(u / 2)
        )
        y_proj = (1j / np.sqrt(2)) * (
            np.exp(-1j * v) * np.cos(u / 2) - np.exp(1j * v) * np.sin(u / 2)
        )

        # -- projections of polarization state |Ψ⟩ on |σ+⟩, |σ-⟩ and |π⟩
        # >>> see documentation for explanation
        # shorthands
        sinB = np.sin(beta)
        cosB = np.cos(beta)
        sinA = np.sin(alpha)
        cosA = np.cos(alpha)
        sq2 = np.sqrt(2)
        # |σ+⟩
        sigma_plus_proj = (cosB * cosA + 1j * sinB) / sq2 * x_proj
        sigma_plus_proj += (sinB * cosA - 1j * cosB) / sq2 * y_proj
        # |σ-⟩
        sigma_minus_proj = (cosB * cosA - 1j * sinB) / sq2 * x_proj
        sigma_minus_proj += (sinB * cosA + 1j * cosB) / sq2 * y_proj
        # |π⟩
        pi_proj = cosB * sinA * x_proj + sinB * sinA * y_proj

        # -- result
        polar_amp = np.array([pi_proj, sigma_plus_proj, sigma_minus_proj]).T

        return polar_amp

    def get_polarization_quant(
        self, quantization_axis: np.ndarray, nocheck: bool = False
    ) -> np.ndarray:
        """Returns **squared norm** the projection of the polarization state |Ψ⟩ on |σ+⟩, |σ-⟩ and |π⟩, using
        the vector ``quantization_axis`` as a quantification axis. See documentation
        for a derivation of this projection.

        Parameters
        ----------
        quantization_axis : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the quantization axis vector in the lab frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        polar_norm : array of shape (3,) or (n1, n2, ..., 3)
            contains the polarization norm for π, σ+ and σ- components

        Notes
        -----

        The input ``quantization_axis`` should be an array of shape (3,) or (n1, n2, .., 3), where
        the cartesian coordinates of the quantization axis are stored in the last dimension (of size 3)

        the result ``polar_norm`` is an array whose size matches the one of ``quantization_axis``, where the last
        dimension of size 3 contains the projections of the polarization state on π, σ+ and σ-

        That is :

        >>> pi_amp, sigmaplus_amp, sigma_minus_amp = polar_norm.T

        With:

        | ``pi_amp`` =  | 〈Ψ|π⟩ | ** 2
        | ``sigmaplus_amp`` =  | 〈Ψ|σ+⟩ | ** 2
        | ``sigma_minus_amp`` =  | 〈Ψ|σ-⟩ | ** 2

        See Also
        --------
        get_polarization_quant_amplitude()
        get_polarization_quant_amplitude_dict()
        get_polarization_quant_dict()
        """

        polar_amp = self.get_polarization_quant_amplitude(quantization_axis, nocheck)
        polar_norm = np.abs(polar_amp) ** 2
        return polar_norm

    def get_polarization_quant_amplitude_dict(
        self, quantization_axis: np.ndarray
    ) -> dict:
        """Returns the projection of the polarization state |Ψ⟩ on |σ+⟩, |σ-⟩ and |π⟩, using
        the vector ``quantization_axis`` as a quantification axis. See documentation
        for a derivation of this projection.

        Parameters
        ----------
        quantization_axis : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the quantization axis vector in the lab frame

        Returns
        -------
        res : dict
            dict containing the polarization amplitude for π, σ+ and σ- components

        Notes
        ------

        The result is returned as a dictionnary `res`, such as :

        | ``res["sigma+"]`` =  〈Ψ|σ+⟩
        | ``res["sigma-"]`` =  〈Ψ|σ-⟩
        | ``res["pi"]`` =  〈Ψ|π⟩

        See Also
        --------
        get_polarization_quant_amplitude()
        get_polarization_quant()
        get_polarization_quant_dict()
        """
        # -- get result in array form
        polar_amp = self.get_polarization_quant_amplitude(quantization_axis)
        pi_amp, sigma_plus_amp, sigma_minus_amp = polar_amp.T

        # -- result
        res = {"sigma+": sigma_plus_amp, "sigma-": sigma_minus_amp, "pi": pi_amp}

        return res

    def get_polarization_quant_dict(self, quantization_axis):
        """Returns the **squared norm** of projection of the polarization state |Ψ⟩ on |σ+⟩, |σ-⟩ and |π⟩, using
        the magnetic field vector ``quantization_axis`` as a quantification axis. See documentation
        for a derivation of this projection.

        Parameters
        ----------
        quantization_axis : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the quantization axis vector in the lab frame

        Returns
        -------
        res : dict
            dict containing the polarization norm for π, σ+ and σ- components

        Notes
        ------

        The result is returned as a dictionnary `res`, such as :

        | ``res["sigma+"]`` =  | 〈Ψ|σ+⟩ | ** 2
        | ``res["sigma-"]`` =  | 〈Ψ|σ-⟩ | ** 2
        | ``res["pi"]`` =  | 〈Ψ|π⟩ | ** 2

        See Also
        --------
        get_polarization_quant_amplitude()
        get_polarization_quant()
        get_polarization_quant_amplitude_dict()

        """
        projection_amplitude = self.get_polarization_quant_amplitude_dict(
            quantization_axis
        )
        res = {}
        for k, v in projection_amplitude.items():
            res[k] = np.linalg.norm(v) ** 2
        return res

    # -- REQUIRED ABSTRACT METHODS

    def get_value(self, position: np.ndarray, nocheck=False) -> np.ndarray:
        """Returns laser intensity at a given position in the lab frame

        Parameters
        ----------
        position : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates in the lab frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        intensity : float or array of shape (n1, n2, ..., 1)
            laser intensity at position

        Notes
        -------
        position is an array_like object, with shape (3,) or (n1, n2, .., 3).
        In all cases, the last dimension contains cordinates (x, y, z), in meter and in the lab frame

        """
        # Check position
        position = self._check_position_array(position, nocheck)
        # call hidden function that actually does the computation
        return self._intensity_func(self, position)

    @abstractmethod
    def _intensity_func(self, position):
        """Actual method for field computation ; defined for each subclass"""

    @abstractmethod
    def set_power_from_I(self, target_I: float):
        """Sets the power to reach a target intensity

        Parameters
        ----------
        target_I : float
            target intensity (W/m^2)
        """

    @abstractmethod
    def set_waist_from_I(self, target_I: float):
        """Sets the waist radius to reach a target intensity

        Parameters
        ----------
        target_I : float
            target intensity (W/m^2)
        """

    # -- CLASS PROPERTIES GETTERS & SETTERS
    # - wavelength
    @property
    def wavelength(self) -> float:
        """float: laser vacuum wavelength (m)"""
        return self._wavelength

    @wavelength.setter
    def wavelength(self, value: float) -> None:
        self._positive_float_check("wavelength", value)
        if value > 3e-6 or value < 100e-9:
            raise Warning(
                "Value given for wavelength is outside the 100nm-3µm range, which is rather strange. Check that you have given the wavelength value in _meters_"
            )
        self._wavelength = float(value)

    # - waist
    @property
    def waist(self) -> float:
        """float: laser 1/e^2 radius (m)"""
        return self._waist

    @waist.setter
    def waist(self, value: float) -> None:
        self._positive_float_check("waist", value)
        self._waist = float(value)

    # - power
    @property
    def power(self) -> float:
        """float: laser power (W)"""
        return self._power

    @power.setter
    def power(self, value: float) -> None:
        self._positive_float_check("power", value)
        self._power = float(value)

    # - waist position
    @property
    def waist_position(self) -> np.ndarray:
        """array of shape (,3): cartesian coordinates of the laser focus / waist position in the lab frame (m)"""
        return self._waist_position

    @waist_position.setter
    def waist_position(self, value: np.ndarray) -> None:
        value = np.asanyarray(value)
        if value.size != 3:
            raise ValueError("'waist_position' should be an array-like of size 3")
        self._waist_position = value

    # - direction_type
    @property
    def direction_type(self) -> str:
        """str: type of direction setting. Can be "vector" or "thetaphi" """
        return self._direction_type

    @direction_type.setter
    def direction_type(self, value: str) -> None:
        if value not in DIRECTION_TYPES:
            raise ValueError(f"'direction_type' should be in {DIRECTION_TYPES}")
        self._direction_type = value

    # - direction
    @property
    def direction(self) -> np.ndarray:
        """array: either a vector or a (theta, phi) tuple describing the laser direction"""
        return self._direction

    @direction.setter
    def direction(self, value: np.ndarray) -> None:
        # convert to array
        value = np.asanyarray(value)

        # check that the size is OK
        errormsg = "When 'direction_type' is set to '{direction_type}', 'direction' should be an array of size {size}"
        if self.direction_type == "vector" and value.size != 3:
            raise ValueError(errormsg.format(direction_type="vector", size=3))
        elif self.direction_type == "thetaphi" and value.size != 2:
            raise ValueError(errormsg.format(direction_type="thetaphi", size=2))

        # compute unit vector
        if self.direction_type == "vector":
            # first case : a unit vector is provided
            # 1 - normalize
            norm = np.linalg.norm(value)
            if norm == 0:
                raise ValueError("Wrong value for the unit vector: norm is zero")
            unit_vector = value / norm
            # 2 - compute theta and phi
            ux, uy, uz = unit_vector
            theta = np.arctan2(np.sqrt(ux**2 + uy**2), uz)
            phi = np.arctan2(uy, ux)

        elif self.direction_type == "thetaphi":
            # second case : theta and phi are provided
            theta, phi = value
            unit_vector = np.array(
                [
                    np.sin(theta) * np.cos(phi),  # x
                    np.sin(theta) * np.sin(phi),  # y
                    np.cos(theta),  # z
                ]
            )
            pass

        # store
        self._unit_vector = unit_vector
        self._unit_vector_phi = phi
        self._unit_vector_theta = theta
        self._direction = value
        # pre compute some values, for later
        self.__costheta = np.cos(theta)
        self.__sintheta = np.sin(theta)
        self.__cosphi = np.cos(phi)
        self.__sinphi = np.sin(phi)

    # - polarization
    @property
    def polarization(self) -> Polarization:
        """Polarization: laser polarization object"""
        return self._polarization

    @polarization.setter
    def polarization(self, value: Polarization) -> None:
        if not isinstance(value, Polarization):
            msg = "`polarization` should be a Polarization object, from atomsmltr.environment.lasers.polarization"
            raise TypeError(msg)
        self._polarization = value

    # - others
    @property
    def unit_vector(self) -> np.ndarray:
        """array of shape (,3): unit vector describing laser propagation"""
        return self._unit_vector

    @property
    def k(self) -> float:
        """float: laser wavenumber k = 2π / λ (m^-1)"""
        return 2 * np.pi / self.wavelength

    @property
    def kvec(self) -> np.ndarray:
        """array: vector version of the laser wavenumber k = 2π / λ (m^-1)"""
        return self.k * self.unit_vector

    # -- hidden methods
    def _positive_float_check(self, param_name: str, value: float) -> None:
        """internal function to check that a parameter is a positive float, raises a `ValueError` if not.

        Args:
            param_name (str): name of the checked parameter, to give context in the exception
            value (float): value of the paramater to check
        """
        if isinstance(value, int):
            value = float(value)
        if not isinstance(value, float):
            raise ValueError(f"'{param_name}' has to be a float")
        if value < 0:
            raise ValueError(f"'{param_name}' has to be a positive")

    # -- PLOT FUNCTIONS

    def plot1D(self):
        pass

    def plot2D(
        self,
        limits: np.ndarray,
        Npoints: np.ndarray,
        cut: float = 0,
        ax=None,
        plane: str = "XY",
        cmap=None,
        show: bool = False,
        space_scale: float = 1.0,
    ):
        """Plots a 2D cut of the laser intensity, using Matplotlib pcolormesh()

        Parameters
        ----------
        limits : array, shape (4,)
            an array of size 4, providing (xmin, xmax, ymin, ymax).
        Npoints : int or array of shape (2,)
            number of points for each dimension,
            either a int or an array of two ints (Nx, Ny).
        cut : float, optional
            coordinate of the third axis for the cut. Defaults to 0.
        ax : Matplotlib Axes, optional
            the matplotlib axis on which to plot.
            If None is given a new figure is created.
            Defaults to None.
        plane : str, optional
            the plane for the cut. Accepted values are "XY", "YZ" and "ZX". Defaults to "XY".
        cmap : Matplotlib cmap, optional
            passed to matplotlib pcolormesh() function
        show : bool, optional
            whether to show the figure after calling the method. Defaults to False.
        space_scale : float, optional
            space coordinates will be multiplied by this when plotting. Defaults to 1.

        Returns
        -------
        ax : Matplotlib Axes
            the axis on which the plot was performed.

        Notes
        ------
        The limits are given via an array of size 4 'limits', providing providing (xmin, xmax, ymin, ymax)
        Number of points are given with 'Npoints', either as an integer (same value for x and y) or an array of size 2
        the coordinate of the cut axis is given by 'cut'

        Examples
        ---------
        >>> beam.plot2D(limits=(-5, 5, -4, 4), Npoints=200)
        >>> beam.plot2D(limits=(-5, 5, -4, 4), Npoints=200, cut=-5)
        >>> beam.plot2D(limits=(-5, 5, -4, 4), Npoints=(200, 100))

        """
        # - process arguments using the Plottable builtin method
        ax, position, X, Y = self._process_2D_plot_args(
            ax=ax,
            plane=plane,
            limits=limits,
            Npoints=Npoints,
            cut=cut,
        )

        # - compute intensity
        intensity = self.get_value(position)

        # - plot
        ax.pcolormesh(X * space_scale, Y * space_scale, intensity, cmap=cmap)
        ax.set_xlabel(plane.upper()[0])
        ax.set_ylabel(plane.upper()[1])

        # - show ?
        if show:
            plt.show()

        return ax

    def plot3D(
        self,
        ax=None,
        color: str = None,
        name: str = None,
        vscale: float = None,
        show: bool = False,
    ):
        """plots a 3D reprensentation of the laser beam, including:
               - a line : laser axis
               - an arrow along the propagation direction
               - a point : laser focus position
               - a dotted arrow : laser polarization vector

        Parameters
        ----------
        ax : custom Axes3D, optional
            the axis in which to plot. If None is given (default value) a new ax is generated
        color : str, optional
            a matplotlib compatible color. Defaults to None.
        name : str, optional
            the name of the laser, passed as a label when plotting. If none is given, use the laser tag
        vscale : float, optional
            A scaling factor. Use it to tweak the arrow size if needed. Defaults to None.
        show : bool, optional
            Whether the show the figure after calling the method. Defaults to False.

        Returns
        -------
        ax : custom Axes3D
            the figure axis in which the laser is plotted.

        Note
        ----
            When providing an axis via the ``ax`` parameter, make sure to use our custom implementation of
            matplotlib ``Axes3D``, as this function uses custom arrow drawing methods. The class can be imported
            via ``from atomsmltr.utils.plotter import Axes3D``
        """
        # - init ax (if needed)
        ax = self._init_ax(ax, ax3D=True)

        # - get laser information
        unit_vector = np.asanyarray(self._unit_vector)
        polar_vector_laserframe = np.asanyarray(self.polarization.vector)
        polar_vector = self._convert_vector_to_lab_frame(polar_vector_laserframe)
        waist_position = np.asanyarray(self.waist_position)
        # - scale
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        zmin, zmax = ax.get_zlim()
        dr = np.array([xmax - xmin, ymax - ymin, zmax - zmin])
        if vscale is None:
            vscale = np.max(dr) / 5

        # - PLOT
        # waist position
        label = self.tag if name is None else name
        ax.scatter(*waist_position, marker="o", color=color, label=label)

        # plot laser
        r1 = waist_position + dr * unit_vector * 5
        r2 = waist_position - dr * unit_vector * 5
        x = np.linspace(-100, 100, 1000)
        r = waist_position[:, np.newaxis] + (unit_vector * dr)[:, np.newaxis] * x
        ax.plot(r[0, :], r[1, :], r[2, :], color=color)

        # plot propagation vector
        epsilon = 0.2
        ax.arrow3D(
            *(waist_position - unit_vector * vscale * (1 + epsilon)),
            *(vscale * unit_vector),
            mutation_scale=15,
            arrowstyle="simple",
            ec="k",
            fc=color,
        )

        # plot polarisation vector
        ax.arrow3D(
            *(waist_position - unit_vector * vscale * (1 + epsilon)),
            *(vscale * polar_vector * 0.7),
            mutation_scale=20,
            arrowstyle="-|>",
            linestyle="dashed",
            color=color,
        )

        if show:
            plt.show()
        return ax

    # -- INFO STRING

    @property
    @abstractmethod
    def disp_type(self) -> str:
        return ""

    def gen_infostring_obj(self, show_polar=True):
        """Generates an info string object"""
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element(f"type", f"{self.disp_type}")
        info.add_element(f"tag", f"{self.tag}")
        info.add_element(f"waist (m)", f"{self.waist:.3g}")
        info.add_element(f"power (W)", f"{self.power:.3g}")
        info.add_element(f"waist position (m)", f"{self.waist_position}")
        info.add_element(f"direction type", f"{self.direction_type}")
        info.add_element(f"direction", f"{self.direction}")
        info.add_element(f"unit vector", f"{self._unit_vector}")
        info.add_element(f"unit vector phi", f"π × {self._unit_vector_phi / np.pi}")
        info.add_element(f"unit vector theta", f"π × {self._unit_vector_theta / np.pi}")

        if show_polar:
            info_polar = self.polarization.gen_infostring_obj()
            info.merge(info_polar, prefix="")

        return info

    def print_info(self, show_polar=True):
        info_str = self.gen_infostring_obj(show_polar)
        print(info_str.generate())

    def print_polar_proj(self, mag_field_vector):
        mag_field_vector = self._check_position_array(mag_field_vector)
        res = self.get_polarization_quant_dict(mag_field_vector)
        print("> Local polarization projection")
        print(f"   + B  = {mag_field_vector*1e4} (G)")
        print(f"   + π  = {res["pi"]:.2f}")
        print(f"   + σ+ = {res["sigma+"]:.2f}")
        print(f"   + σ- = {res["sigma-"]:.2f}")


# % IMPLEMENTED CLASSES


class GaussianLaserBeam(LaserBeam):
    """A Gaussian laser beam

    Parameters
    ----------
    wavelength : float, optional
        vacuum wavelength (m), by default 399e-9
    waist : float, optional
        1/e^2 waist radius (m), by default 1e-3
    power : float, optional
        laser power (W), by default 1e-3
    waist_position : array, shape (,3), optional
        cartesian coordinates of the waist / focus position,
        in meters and in the lab frame, by default (0, 0, 0)
    direction : array, shape (,3) or (,2), optional
        depending on 'direction type', a vector or a (theta, phi) couple
        giving the propagation direction of the beam
    direction_type : str, optional
        type of direction : "vector" or "thetaphi", by default "vector"
    polarization : Polarization, optional
        laser polarization, by default Vertical()
    tag : str, optional
        laser tag, by default None
    """

    @property
    def type(self):
        return "Gaussian Laser Beam"

    @property
    def disp_type(self) -> str:
        return "Gaussian beam"

    # -- REQUIRED METHOD FOR LASER BEAM CLASSES
    # pylint : disable=method_hidden
    @staticmethod
    def _intensity_func(self, position):
        """Returns laser intensity at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `LaserBeam` class
        """
        # - get coordinates in laser frame
        # NB : x, y and phi are not needed here
        position_laser = self._convert_coordinates_to_laser_frame(position)
        x_laser, y_laser, z_laser = position_laser.T
        rho_laser = np.sqrt(x_laser**2 + y_laser**2)
        # - compute gaussian beam intensity
        intensity = _intensity_gauss(
            rho_laser, z_laser, self.waist, self.power, self.wavelength
        )
        intensity = intensity.T
        return intensity

    def set_power_from_I(self, target_I: float) -> None:
        # NB, for a Gaussian beam : I0 = 2 * P / np.pi / w0**2
        power = target_I * self.waist**2 * np.pi / 2
        self.power = power

    def set_waist_from_I(self, target_I: float) -> None:
        # NB, for a Gaussian beam : I0 = 2 * P / np.pi / w0**2
        waist = np.sqrt(2 * self.power / np.pi * target_I)
        self.waist = waist

    @property
    def rayleigh_length(self) -> float:
        """float: the laser Rayleigh length"""
        return np.pi * self.waist**2 / self.wavelength

    def gen_infostring_obj(self, show_polar=True):
        info = super().gen_infostring_obj(show_polar)
        info.add_element(
            "Rayleigh length", f"{self.rayleigh_length:.2g} m", section="Parameters"
        )
        return info


class PlaneWaveLaserBeam(LaserBeam):
    """Implements a plane wave. For convenience, we still define the waist and power, and
    the intensity is constant and corresponds to the peak intensity of a Gaussian beam with
    same power and waist."""

    @property
    def type(self):
        return "Plane Wave Laser Beam"

    @property
    def disp_type(self) -> str:
        return "Plane wave beam"

    # -- REQUIRED METHOD FOR LASER BEAM CLASSES
    # pylint : disable=method_hidden
    @staticmethod
    def _intensity_func(self, position: np.ndarray) -> np.ndarray:
        """Returns laser intensity at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `LaserBeam` class
        """
        # - get coordinates in laser frame
        # NB : x, y and phi are not needed here
        x, _, _ = position.T
        x = x.T
        # - compute gaussian beam intensity
        intensity = _intensity_gauss(
            0 * x, 0 * x, self.waist, self.power, self.wavelength
        )

        return intensity

    def set_power_from_I(self, target_I: float) -> None:
        # NB, for a Gaussian beam : I0 = 2 * P / np.pi / w0**2
        power = target_I * self.waist**2 * np.pi / 2
        self.power = power

    def set_waist_from_I(self, target_I: float) -> None:
        # NB, for a Gaussian beam : I0 = 2 * P / np.pi / w0**2
        waist = np.sqrt(2 * self.power / np.pi * target_I)
        self.waist = waist

    def gen_infostring_obj(self, show_polar=True):
        info = super().gen_infostring_obj(show_polar)
        info.rm_element("waist position (m)", section="Parameters")
        return info



"""The ``atomsmltr.atoms.collection`` subpackage provides definitions for common atomic species

Examples
---------

Import an ytterbium atom

.. code-block:: python

    from atomsmltr.atoms import Ytterbium  # import `Ytterbium` class
    yb = Ytterbium()  # init an object
    yb.print_info()  # print atom informations

"""

__all__ = [
    "Ytterbium",
    "Strontium",
    "Rubidium",
]

from .ytterbium import Ytterbium
from .strontium import Strontium
from .rubidium import Rubidium



"""simulator
==================

Here we implement the ``Simulation`` class, that allows to run simulations
on a configuration defined via the ``Configuration`` class.

Examples
--------------------

Run a simulation with one initial condition vector

.. code-block:: python

        # ... init a config object with the `Configuration` class

        # - import a simulation class
        from atomsmltr.simulation import ScipyIVP_3D

        # - init and setup
        sim = ScipyIVP_3D(method="Radau")
        sim.config = config

        # - parameters
        u0 = (0, 0, -0.15, 0, 0, 200)
        t = np.linspace(0, 0.05, 1000)

        # - integrate
        res = sim.integrate(u0, t)

Run a batch of simulations

.. code-block:: python

        # ... init a config object with the `Configuration` class

        # - import a simulation class
        from atomsmltr.simulation import ScipyIVP_3D

        # - init and setup
        sim = ScipyIVP_3D(method="Radau")
        sim.config = config

        # - parameters
        # initial conditions
        vz_list = np.linspace(10, 300, 40)
        u0_list = [(0, 0, -0.15, 0, 0, v) for v in vz_list]
        sim.u0_list = u0_list
        # time
        t = np.linspace(0, 0.05, 1000)

        # - run a batch in parallel
        res_list = sim.run(t, npools=5, verbose=True)

"""

__all__ = [
    "Simulation",
    "ScipyIVP_3D",
    "RK4",
    "RK4St",
    "Euler",
    "EulerSt",
    "VelocityVerlet",
]

from .simbase import Simulation
from .scipy import ScipyIVP_3D
from .deterministic import RK4, Euler, VelocityVerlet
from .stochastic import RK4St, EulerSt



"""
zones
=======================

Here we implement the generic ``Zone`` class, as well as a series of
``ZoneCollection`` classes that are used to combine several zones

Note
----
    the actual implementation of zones are in other modules

See Also
--------
atomsmltr.environment.zones.limits
atomsmltr.environment.zones.volumes

"""

# % IMPORTS
import numpy as np
from abc import abstractmethod
from copy import copy, deepcopy

# % LOCAL IMPORTS
from ..envbase import EnvObject
from ...utils.misc import check_position_speed_array
from ...utils.infostring import InfoString

# % ABSTRACT CLASSES

IMPLEMENTED_ACTIONS = ["stop", "ignore"]
IMPLEMENTED_TARGETS = ["position", "speed"]


# % -------------------------------
# % SIMPLE ZONES
# % -------------------------------


class Zone(EnvObject):
    """A generic Zone object

    Parameters
    ----------
    target : str, optional
        the target for the zone, can be "position" or "speed", by default "position"
    action : str, optional
        the action associated to the zone.
        implemented actions = ["stop", "ignore"], default is "stop"
    tag : str, optional
        the zone tag
    in_tag : str, optional
        tag for an object inside the zone, by default None
    out_tag : str, optional
        tag for an object inside the zone, by default None
    """

    def __init__(
        self,
        target: str = "position",
        action: str = "stop",
        tag: str = None,
        in_tag: str = None,
        out_tag: str = None,
    ):
        super(Zone, self).__init__(tag)
        self.inverted = False
        self.target = target
        self.action = action
        self.in_tag = in_tag
        self.out_tag = out_tag

    # -- GETTERS & SETTERS

    @property
    def vector(self):
        return False

    @property
    def inverted(self):
        """bool: if inverted, the zone logic is inverted"""
        return self.__inverted

    @inverted.setter
    def inverted(self, value):
        if not isinstance(value, bool):
            raise TypeError("'inverted' should be a boolean")
        self.__inverted = value

    @property
    def target(self):
        """str: the target for the zone. Can be "position" or "speed" """
        return self.__target

    @target.setter
    def target(self, value):
        if value not in IMPLEMENTED_TARGETS:
            raise ValueError(f"implemented targets are : {IMPLEMENTED_TARGETS}")
        self.__target = value

    @property
    def action(self):
        """str: the action associated with the zone. implemented actions = ["stop", "ignore"]."""
        return self.__action

    @action.setter
    def action(self, value):
        if value not in IMPLEMENTED_ACTIONS:
            raise ValueError(f"implemented actions are : {IMPLEMENTED_ACTIONS}")
        self.__action = value

    @property
    def in_tag(self) -> str:
        """str: tag for a object inside the zone"""
        return self._in_tag

    @in_tag.setter
    def in_tag(self, value: str) -> None:
        if not isinstance(value, str) and value is not None:
            raise TypeError("'in_tag' should be a string or None")
        self._in_tag = value

    @property
    def out_tag(self) -> str:
        """str: tag for a object outside the zone"""
        return self._out_tag

    @out_tag.setter
    def out_tag(self, value: str) -> None:
        if not isinstance(value, str) and value is not None:
            raise TypeError("'out_tag' should be a string or None")
        self._out_tag = value

    # -- functions

    def invert(self):
        """toggles the 'inverted' status"""
        self.__inverted = not self.__inverted

    def inverted_copy(self):
        """Returns an inverted copy of the object"""
        new_object = deepcopy(self)
        new_object.invert()
        return new_object

    # -- METHODS
    def get_value(self, vector: np.ndarray, nocheck: bool = False) -> np.ndarray:
        """Evaluates whether 'vector' is in the zone

        Parameters
        ----------
        vector : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates of the vectors in the lab frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        value : array of shape (1,) or (n1, n2, ..., 1)
            wheter the vector is 'in the zone'

        Notes
        -----

        ``vector`` should be an array of shape (...,3), where last axis contains
        the coordinates to evaluate.

        if the ``inverted`` property is set to true, ``get_value`` will return
        True outside the zone

        """
        vector = self._check_position_array(vector, nocheck)
        res = self._in_zone(vector)
        if self.inverted:
            res = np.logical_not(res)
        return res

    @abstractmethod
    def _in_zone(self, vector):
        """actual implementationf of 'get_value'"""

    # -- OPERATORS OVERLOADING

    def __and__(self, object):
        if isinstance(object, Zone):
            new_collection = ANDCollection()
            new_collection.add_zone(deepcopy(self))
            new_collection.add_zone(deepcopy(object))
            return new_collection
        else:
            raise TypeError("only 'Zones' objects can be combined")

    def __or__(self, object):
        if isinstance(object, Zone):
            new_collection = ORCollection()
            new_collection.add_zone(deepcopy(self))
            new_collection.add_zone(deepcopy(object))
            return new_collection
        else:
            raise TypeError("only 'Zones' objects can be combined")

    def __xor__(self, object):
        if isinstance(object, Zone):
            new_collection = XORCollection()
            new_collection.add_zone(deepcopy(self))
            new_collection.add_zone(deepcopy(object))
            return new_collection
        else:
            raise TypeError("only 'Zones' objects can be combined")


# % -------------------------------
# % ZONES COLLECTIONS
# % -------------------------------


class ZoneCollection(Zone):
    def __init__(self, *args, **kwargs):
        self.__zones = []
        super(ZoneCollection, self).__init__(*args, **kwargs)

    # -- METHODS AND PROPERTIES
    @property
    def type(self):
        return "Zone Collection"

    @property  # readonly
    def zones(self) -> list:
        """list: a list of the zones included in the collection"""
        return self.__zones

    def add_zone(self, zone: Zone):
        """adds a zone to the current collection

        Parameters
        ----------
        zone : Zone
            the zone to add
        """

        if not isinstance(zone, Zone):
            raise TypeError("'zone' should be a zone object")
        self.__zones.append(zone)

    def reset(self):
        """Resets the zone list"""
        self.__zones = []

    # -- INFOSTRING

    def gen_infostring_obj(self):
        """Generates an info string object"""
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", self.type)
        info.add_element("tag", self.tag)
        info.add_element("in_tag", self.in_tag)
        info.add_element("out_tag", self.out_tag)
        info.add_element("target", self.target)
        info.add_element("action", self.action)
        info.add_element(f"zones", f"{[z.tag for z in self.zones]}")
        info.add_element(f"inverted", f"{self.inverted}")
        return info

    # -- PLOT

    def plot1D(self, ax=None):
        pass

    def plot2D(self, ax=None):
        pass

    def plot3D(self, ax=None):
        pass

    # -- OPERATORS OVERLOADING

    def __add__(self, object):
        # then operator acts on a new collection
        collection = self.__class__()
        for z in self.zones:
            collection.add_zone(deepcopy(z))
        return self.__add_operator__(object, collection)

    def __iadd__(self, object):
        """let's handle additions between zonecollections

        add behaves as a shorthand for "add_zones"
        we will only allow collections of same type to be added

        """
        # then operator acts on self
        # - recursive add if list
        if isinstance(object, (list, tuple)):
            for element in object:
                self.__add_operator__(element, self)
            return self
        # - otherwise
        return self.__add_operator__(object, self)

    def __add_operator__(self, object, coll):
        """a function to factor the __add__ and __iadd__ operators"""
        # case 1 > same type of zone, then we add all the zones
        if object:
            if isinstance(object, self.__class__):
                for z in object.zones:
                    new_zone = deepcopy(z)
                    if object.inverted:
                        new_zone.invert()
                    coll.add_zone(new_zone)
                return coll
            # case 2 > it is a zone, not a collection
            elif isinstance(object, Zone) and not isinstance(object, ZoneCollection):
                coll.add_zone(deepcopy(object))
                return coll
            else:
                raise TypeError(
                    "a ZoneCollection can only be added with a Zone or another ZoneCollection of same type"
                )


class ANDCollection(ZoneCollection):

    # -- METHODS AND PROPERTIES
    @property
    def type(self):
        return "AND Zone Collection"

    def _in_zone(self, vector):
        res_list = [zone.get_value(vector) for zone in self.zones]
        return np.logical_and.reduce(res_list)


class ORCollection(ZoneCollection):

    # -- METHODS AND PROPERTIES
    @property
    def type(self):
        return "OR Zone Collection"

    def _in_zone(self, vector):
        res_list = [zone.get_value(vector) for zone in self.zones]
        return np.logical_or.reduce(res_list)


class XORCollection(ZoneCollection):

    # -- METHODS AND PROPERTIES
    @property
    def type(self):
        return "XOR Zone Collection"

    def _in_zone(self, vector):
        res_list = [zone.get_value(vector) for zone in self.zones]
        return np.logical_xor.reduce(res_list)


# % -------------------------------
# % SUPER ZONE
# % -------------------------------

IMPLEMENTED_LOGIC = ["OR", "XOR", "AND"]


class SuperZone(ZoneCollection):
    """SuperZone is a zone collection for position/speed vectors

    Parameters
    ----------
    zones : list, optional
        list of zones to add at object creation, by default []
    logic : str, optional
        the logic of zone combination. Can be "OR", "AND", "XOR", by default "AND"
    action : str, optional
        the action to trigger
        implemented actions = ["stop", "ignore"], default is "stop"
    tag : str, optional
        the tag of the zone, by default None
    in_tag : str, optional
        tag for an object inside the zone, by default None
    out_tag : str, optional
        tag for an object inside the zone, by default None
    """

    def __init__(
        self,
        zones: list = [],
        logic: str = "AND",
        action: str = "stop",
        tag: str = None,
        in_tag: str = None,
        out_tag: str = None,
    ):

        super(SuperZone, self).__init__(
            action=action, tag=tag, in_tag=in_tag, out_tag=out_tag
        )
        self.logic = logic
        self.__iadd__(zones)

    # -- PROPERTIES

    @property
    def type(self):
        return "Super Zone collection"

    # - target is irrelevant
    @property
    def target(self):
        """not used in this case"""
        return None

    @target.setter
    def target(self, value):
        pass

    # - logic
    @property
    def logic(self):
        """str: the logic for the SuperZone combination ("OR", "XOR", "AND")"""
        return self.__logic

    @logic.setter
    def logic(self, value: str):
        if value not in IMPLEMENTED_LOGIC:
            raise ValueError(f"'logic' should be in {IMPLEMENTED_LOGIC}")
        self.__logic = value
        match value:
            case "OR":
                self.__logical_op = np.logical_or
            case "AND":
                self.__logical_op = np.logical_and
            case "XOR":
                self.__logical_op = np.logical_xor

    # -- METHODS

    def add_zones(self, zones: Zone | list):
        self.__iadd__(zones)

    def get_value(self, vector: np.ndarray, nocheck: bool = False) -> np.ndarray:
        """Evaluates whether 'vector' is in the zone

        Parameters
        ----------
        vector : array of shape (6,) or (n1, n2, ..., 6)
            cartesian coordinates of the vectors in the lab frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        value : array of shape (1,) or (n1, n2, ..., 1)
            whether the vector is 'in the zone'

        Notes
        -----

        This is a ``SuperZone`` object, so it acts on **position-speed** vectors
        of dimensions 6 !

        ``vector`` should be an array of shape (6,) or (n1, n2, .., 6), where last axis contains
        the coordinates (position & speed) to evaluate.

        In all cases, the last dimension contains cordinates (x, y, z, vx, vy, vz),

        if the ``inverted`` property is set to true, ``get_value`` will return
        True outside the zone

        """
        vector = check_position_speed_array(vector, nocheck)
        res = self._in_zone(vector)
        if self.inverted:
            res = np.logical_not(res)
        return res

    def _in_zone(self, vector: np.ndarray) -> np.ndarray:
        """The actual implementation of the ``in_zone`` method"""
        # -- separate speeds & positions
        x, y, z, vx, vy, vz = vector.T
        position = np.array([x, y, z]).T
        speed = np.array([vx, vy, vz]).T

        # -- separate zones according to target
        speed_zones = []
        position_zones = []
        for zone in self.zones:
            if zone.target == "speed":
                speed_zones.append(zone)
            elif zone.target == "position":
                position_zones.append(zone)

        # -- evaluate
        res_list = [zone.get_value(position) for zone in position_zones]
        res_list += [zone.get_value(speed) for zone in speed_zones]

        if res_list:
            res = self.__logical_op.reduce(res_list)
        else:
            res = x.T == x.T
        return res

    def gen_infostring_obj(self):
        info = super().gen_infostring_obj()
        info.add_element("logic", self.logic)
        info.rm_element("target")
        return info



"""
environment objects
=======================

This module implements the generic ``EnvObject`` class. All objects aimed at defining
the atom's environment will derive from this class.

This will help ensure that those objects share common properties and methods, which will
ease their use inside a ``Configuration`` class.
"""

# % IMPORTS
from abc import abstractmethod
from copy import copy
import numpy as np

# % LOCAL IMPORTS
from ..utils.plotter import Plottable
from ..utils.misc import random_word, check_position_array

# % ABSTRACT CLASSES


class EnvObject(Plottable):
    """Defines a generic (abstract) class for environment objects.

    Notably, this makes sure that:

      (1) they are ``Plottable`` objects
      (2) they feature a ``get_value()`` method that satisfies our vectorization convention
      (3) they have a ``tag`` property.

    Parameters
    ----------
    tag : str, optional (default=None)
        a tag identifying the object. If `None` is given, a random tag
        will be generated

    See Also
    ----------
    atomsmltr.utils.plotter.plotter.Plottable
    """

    def __init__(self, tag: str = None):
        # init tag with random word if None
        if tag is None:
            tag = random_word()
        self.tag = tag
        super(EnvObject, self).__init__()

    # -- UNIFIED GET_VALUE FUNCTION

    @abstractmethod
    def get_value(self, position: np.ndarray, nocheck=False) -> np.ndarray:
        """Returns the ``EnvObject`` 'value' at a given position

        Parameters
        ----------
        position : array of shape (3,) or (n1, n2, ..., 3)
            cartesian coordinates in the lab frame
        nocheck : bool, optional
            if set to True, function will not check that the shape of position
            matches requirements, by default False

        Returns
        -------
        intensity : float or array of shape (n1, n2, ..., 1)
            laser intensity at position

        Notes
        -------
        position is an array_like object, with shape (3,) or (n1, n2, .., 3).
        In all cases, the last dimension contains cordinates (x, y, z), in meter and in the lab frame

        Note
        -----
            this is an **abstract** method, and the actual method has to be implemented
            for each child class.
        """

        # -- check position array
        pass

    def _check_position_array(self, position: np.ndarray, nocheck=False) -> np.ndarray:
        """Checks that a position array matches our vectorization convention.

        It raises an error if the shape is not good.

        Parameters
        ----------
        position : array
            the array to check
        nocheck : bool, optional
            if set to True, the function is bypasses

        Returns
        -------
        position
            the array

        Notes
        ------
        positions array should have a shape (3,) or (n1, n2, .., 3).

        In all cases, the last dimension contains cordinates (x, y, z),
        in meter and in the lab frame
        """
        return check_position_array(position, nocheck)

    # -- TAG
    @property
    def tag(self) -> str:
        """str: the object tag"""
        return self._tag

    @tag.setter
    def tag(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("'tag' should be a string")
        self._tag = value

    # -- COPY
    def copy(self, new_tag: str = None):
        """returns a copy of the object.

        Parameters
        ----------
        new_tag : str, optional
            a tag for the copy, if None is given a random tag is generated.

        Returns
        -------
        object_copy : EnvObject
            a copy of the object
        """
        object_copy = copy(self)
        if new_tag is None:
            new_tag = f"{self.tag}-{random_word()}"
        object_copy.tag = new_tag
        return object_copy

    # -- INFO STRING / OBJECT MANAGEMENT
    @abstractmethod
    def gen_infostring_obj(self):
        """generates an ``InfoString`` object.

        Returns
        -------
        InfoString
            an ``InfoString`` object

        See also
        --------
        atomsmltr.utils.infostring.InfoString
        """
        pass

    def gen_info_string(self, **kwargs):
        """generates an info string

        Returns
        -------
        info_string: str
            a string with information on the atom
        """
        return self.gen_infostring_obj().generate(**kwargs)

    def print_info(self):
        """prints the atom infostring"""
        print(self.gen_info_string())

    @property
    @abstractmethod
    def type():
        """str: a description of the object type"""
        pass

    @property
    @abstractmethod
    def vector():
        """bool: True it the object value is vectorial, False if scalar"""
        pass



"""simulators objects
=======================

Here we implements the generic ``Simulation`` class. Actual implementions are to be
defined in other submodules.
"""

# % IMPORTS
import numpy as np
from scipy import constants as csts
from abc import ABC, abstractmethod
from multiprocessing import Pool
from tqdm import tqdm
from functools import partial
from dataclasses import dataclass

# % LOCAL IMPORTS
from ..configurator import Configuration

# % USEFUL FUNCTIONS


def _get_force_vec(
    position: np.ndarray,
    speed: np.ndarray,
    config: Configuration,
    return_list: bool = False,
) -> np.ndarray:
    """Computes the force on an atom, by adding all radiations pressures,
    in a vectorization style that matches the package's standards

    Parameters
    ----------
    position : array, shape (3,) or (n1, n2, .., 3)
        array of cartesian positions in the lab frame
    speed : array, shape (3,) or (n1, n2, .., 3)
        array of cartesian speeds in the lab frame
    config : Configuration
        a configuration object
    return_list : bool (opt, default=False)
        if set to True, also returns the scattering rates and laser wavenumbers
        this is used to compute the stochastic part of the force (spont. em.)

    Returns
    -------
    force : array, (3,) or (n1, n2, .., 3)
        the force felt by the atoms

    scattering_list : list
        returned only if ``return_list`` is set to ``True``
        list of lasers scattering rates and wavenumbers.
    """

    # - get magnetic field value & norm
    B = config.getB(position)
    Bx, By, Bz = B.T
    B_norm = np.sqrt(Bx**2 + By**2 + Bz**2).T
    # - initialize force
    force = np.zeros_like(position, dtype=float)
    # - prepare scattering list
    scattering_list = []
    # - loop over atom-light couplings
    atomlight_couples = config.get_atomlight_couples()
    for elements in atomlight_couples:
        transition, laser, detuning = elements
        laser_intensity = laser.get_value(position)
        polarization = laser.get_polarization_quant(B)
        # Doppler
        det_Doppler = -np.dot(speed, transition.k * laser.unit_vector)
        scattering_rate = transition.get_scattering_rate(
            laser_intensity, B_norm, polarization, detuning + det_Doppler
        )
        radiation_pressure = csts.hbar * transition.k * scattering_rate
        force = force + radiation_pressure[..., np.newaxis] * laser.unit_vector
        if return_list:
            scattering_list.append(
                {
                    "rate": scattering_rate,
                    "k": transition.k,
                    "unit_vector": laser.unit_vector,
                }
            )

    # - loop over all forces
    for f in config.get_all_forces():
        force = force + f.get_value(position)

    if return_list:
        return force, scattering_list
    else:
        return force


def get_force_vec(
    pos_speed_vector: np.ndarray,
    config: Configuration,
    return_list: bool = False,
) -> np.ndarray:
    """Computes the force on an atom, by adding all radiations pressures,
    in a vectorization style that matches the package's standards

    Parameters
    ----------
    pos_speed_vector : array, shape (6,) or (n1, n2, .., 6)
        array of cartesian coordinates (position and speed) in the lab frame
    config : Configuration
        a configuration object
    return_list : bool (opt, default=False)
        if set to True, also returns the scattering rates and laser wavenumbers
        this is used to compute the stochastic part of the force (spont. em.)

    Returns
    -------
    force : array, shape (3,) or (n1, n2, .., 3)
        the force at the coordinates given by ``pos_speed_vector``

    scattering_list : list
        returned only if ``return_list`` is set to ``True``
        list of lasers scattering rates and wavenumbers.

    Notes
    -----
    ``pos_speed_vector`` is an array_like object, with shape (6,) or (n1, n2, .., 6).

    In all cases, the last dimension contains cordinates (x, y, z, vx, vy, vz),
    in meter or meter/seconds and in the lab frame

    Examples
    --------

    .. code-block:: python

        # ... init a config object with the `Configuration` class
        from atomsmltr.simulation.simulator import get_force_vec
        import numpy as np

        # - init a position & speed vector grid
        # x spans from -0.1 to 0.1
        # vx spans from -10 to 30
        # y, z, vy, vz set to 0
        grid = np.mgrid[
            -0.1:0.1:100j,  # x
                0:0:1j,  # y
                0:0:1j,  # z
            -10:30:101j,  # vx
                0:0:1j,  # vy
                0:0:1j,  # vz
        ]
        # squeeze unused dimensions
        grid = np.squeeze(grid)
        # get X and VX grids (for instance for plotting)
        X, _, _, VX, _, _ = grid
        # make (x, y, z, vx, vy, vz) the last dimension
        # as requested by vectorization convention
        pos_speed_vector = grid.T

        # - compute the force
        force = get_force_vec(pos_speed_vector, config)
        FX, FY, FZ = force.T

        # - print shapes for illustration
        print(f"{grid.shape=}")
        print(f"{X.shape=}")
        print(f"{FX.shape=}")
        print(f"{pos_speed_vector.shape=}")
        print(f"{force.shape=}")


    This returns

    .. code-block:: python

        grid.shape=(6, 100, 101)
        X.shape=(100, 101)
        FX.shape=(100, 101)
        pos_speed_vector.shape=(101, 100, 6)
        force.shape=(101, 100, 3)
    """

    # TODO should we move that to the Configuration class ???
    # - get position and speed
    x, y, z, vx, vy, vz = pos_speed_vector.T
    position = np.array([x, y, z]).T
    speed = np.array([vx, vy, vz]).T
    # - get force
    res = _get_force_vec(position, speed, config, return_list)
    if return_list:
        force, scattering_list = res
        return force, scattering_list
    else:
        force = res
        return force


# % DEFINE THE BASE CLASS


@dataclass
class SimRes:
    """Class for simulation results"""

    y: np.ndarray
    t: np.ndarray
    y_last: np.ndarray = None
    tags: set = None
    t_events: list = None
    y_events: list = None
    success: bool = True


class Simulation(ABC):
    """The generic Simulation object

    Parameters
    ----------
    config : Configuration, optional
        the configuration to consider, by default None

    Note
    -----
        this is an abstract class, actual implementations are
        defined elsewhere and inherit from this class
    """

    def __init__(self, config: Configuration = None):
        super(Simulation, self).__init__()
        if config is not None:
            self.config = config
        self.u0_list = []

    # -- SETTERS AND GETTERS
    @property
    def config(self):
        """Configuration: the configuration for this simulation"""
        return self.__config

    @config.setter
    def config(self, value):
        if not isinstance(value, Configuration):
            raise TypeError("'config' should be a `Configuration` object.")
        self.__config = value

    @property
    def u0_list(self):
        """list: a list of initial conditions for batch running"""
        return self.__u0_list

    @u0_list.setter
    def u0_list(self, value):
        value = self._u0_list_checker(value)
        self.__u0_list = value

    # -- REQUESTED FUNCTIONS
    @abstractmethod
    def get_force(self, u: np.ndarray) -> np.ndarray:
        """returns the force felt at a position/speed vector u

        Parameters
        ----------
        u : array, shape (6,) or (n1, n2, .., 6)
            array of cartesian coordinates (position and speed) in the lab frame

        Returns
        -------
        force : array, shape (3,) or (n1, n2, .., 3)
            the force at the coordinates given by ``pos_speed_vector``
        """
        pass

    def integrate(self, u0: np.ndarray, t: np.ndarray):
        """Integrates the system with initial conditions ``u0``

        Parameters
        ----------
        u0 : array, shape (6,)
            the initial conditions (x, y, z, vx, vy, vz)
        t : array, shape (n,)
            the timesteps to integrate

        Returns
        -------
        res
            the result of the simulation
        """
        ####################
        #  PRE PROCESSING  #
        ####################
        # for later use

        #################
        #  INTEGRATION  #
        #################
        #  integrate using the method specific `_integrate()` method
        res = self._integrate(u0, t)

        #####################
        #  POST PROCESSING  #
        #####################
        # - apply zones tags
        # get zones
        position_zones, speed_zones = self.config.get_all_zones()
        # get last position
        # -------------------------------------------------------
        # Note : we have to take into account the case where
        #        u0 is a vector of shape (n, m, ..., 6)
        #        and stop times might be different for all
        #        dimensions. In this case, when one trajectory
        #        is "stopped", it is filled with nan. Thus we
        #        will take all values for u backwards in time,
        #        and replace all nans until we have no nans
        # -------------------------------------------------------
        # 1) take last value
        # since res.y has a shape (n, m, ..., 6, N) where
        # N is the number of timesteps, we transpose to make
        # it easier to iterate on timesteps
        yT = res.y.T
        # take the last time step
        uT_last = yT[-1, ...]
        # iterate backward in time
        for uT in yT[::-1]:
            # we replace the nan values in the current vector
            # by the ones from the last timestep on which we iterate
            # non nan values are kept
            uT_last = np.where(np.isnan(uT_last), uT, uT_last)
            # if we have no nan left, we stop
            if not np.any(np.isnan(uT_last)):
                break
        # transpose it back
        u_last = uT_last.T
        # store it
        res.y_last = u_last
        # extract speed and position
        position = u_last[..., :3]
        speed = u_last[..., 3:]
        res.tags = set()  # we use a set to have unique values
        # add position tags
        for zone in position_zones:
            new_tags = np.where(
                zone.get_value(position),
                {zone.in_tag},
                {zone.out_tag},
            )
            res.tags |= new_tags
        # add speed tags
        for zone in speed_zones:
            new_tags = np.where(
                zone.get_value(speed),
                {zone.in_tag},
                {zone.out_tag},
            )
            res.tags |= new_tags

        return res

    @abstractmethod
    def _integrate(self, u0, t):
        """actual integration"""
        pass

    @abstractmethod
    def dudt(self, t, u):
        """should return the derivative of the position/speed vector u"""
        pass

    @abstractmethod
    def _u0_list_checker(self, value):
        """checks that the list of initial conditions matches what is expected
        for a given simulator implementation"""
        pass

    # -- RUN
    def run(
        self,
        t: np.ndarray,
        u0_list: list = None,
        npools: int = 0,
        verbose: bool = False,
    ) -> list:
        """Runs a batch of simulations from a list of initial conditions

        Parameters
        ----------
        t : array, shape (n,)
            time steps for the simulation
        u0_list : list, optional
            list of initial conditions, by default None
        npools : int, optional
            number of pools for parallel computing.
            If set to zero, no paralalelisation, by default 0
        verbose : bool, optional
            if set to True, a progress bar is displayed, by default False

        Returns
        -------
        res_list : list
            a list of results

        Examples
        --------

        .. code-block:: python

            # ... init a config object with the `Configuration` class

            # - import a simulation class
            from atomsmltr.simulation import ScipyIVP_3D

            # - init and setup
            sim = ScipyIVP_3D(method="Radau")
            sim.config = config

            # - parameters
            # initial conditions
            vz_list = np.linspace(10, 300, 40)
            u0_list = [(0, 0, -0.15, 0, 0, v) for v in vz_list]
            sim.u0_list = u0_list
            # time
            t = np.linspace(0, 0.05, 1000)

            # - run a batch in parallel
            res_list = sim.run(t, npools=5, verbose=True)

        """
        if u0_list is not None:
            self.u0_list = u0_list
        if not isinstance(npools, int):
            return TypeError("'npools' should be an int")
        if npools:
            map_fun = partial(self.integrate, t=t)
            if verbose:
                Nmax = len(self.u0_list)
                res_list = []
                with Pool(npools) as p, tqdm(total=Nmax) as pbar:
                    for res in p.imap(map_fun, self.u0_list):
                        pbar.update()
                        pbar.refresh()
                        res_list.append(res)
            else:
                with Pool(npools) as p:
                    res_list = p.map(map_fun, self.u0_list)
        else:
            res_list = []
            u0_list = tqdm(self.u0_list) if verbose else self.u0_list
            for u0 in u0_list:
                res = self.integrate(u0, t)
                res_list.append(res)
        return res_list



"""
modifiers for environment objects
===================================

This module contains several modifiers (decorators) for environment
objects, that can be used to perform spatial translations or
rotation on those objects.

Warning
--------

(!) This module is still at an experimental stage (!)

Do not use it yet, as it might yield unexpected results


Examples
---------

Rotate a laser beam

.. code-block:: python

    import numpy as np
    from atomsmltr.environment import GaussianLaserBeam
    from atomsmltr.environment.modifiers import rotate

    laser = GaussianLaserBeam(direction=(1,0,0))
    rotate(laser, (0,0,1), np.pi/4)


Shift a magnetic field quadrupole

.. code-block:: python

    from atomsmltr.environment import MagneticQuadrupoleX
    from atomsmltr.environment.modifiers import shift

    mag_field = MagneticQuadrupoleX(origin=(0,0,0), slope=1)
    shift(mag_field, (-1,0,1))

"""

# % IMPORTS
import numpy as np
from functools import wraps

# % LOCAL IMPORTS
from .envbase import EnvObject
from ..utils.misc import check_position_array

# % USEFUL FUNCTIONS


def rotation_matrix(u: np.ndarray, theta: float) -> np.ndarray:
    """Generates 3D rotation matrix

    Parameters
    ----------
    u : array, shape (3,)
        the axis around which to perform the rotation
        it does not need to be normalized, the function will take
        care of it
    theta : float
        the angle for the rotation

    Returns
    -------
    rotmat : array, shape (3,)
        the rotation matrix

    References
    -----------
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle

    """
    # - normalize vector
    u = np.asanyarray(u)
    assert u.shape == (3,), "'u' should be an array of shape (3,)"
    assert np.linalg.norm(u) > 0, "'u' should not be null"
    u = u / np.linalg.norm(u)

    # - prepare useful matrices
    I = np.eye(3)
    u_cross = np.cross(I, u)
    u_out = np.outer(u, u)

    # - generate rotation matrix
    rotmat = np.cos(theta) * I + np.sin(theta) * u_cross + (1 - np.cos(theta)) * u_out

    return rotmat


def rotate_position_vector(
    position: np.ndarray, u: np.ndarray, theta: float
) -> np.ndarray:
    """Rotates a position vector

    Parameters
    ----------
    position : array of shape (3,) or (n1, n2, ..., 3)
        cartesian coordinates in the lab frame
    u : array, shape (3,)
        the axis around which to perform the rotation
        it does not need to be normalized, the function will take
        care of it
    theta : float
        the angle for the rotation

    Returns
    -------
    rotated_position : array of shape (3,) or (n1, n2, ..., 3)
        rotated position vector

    Notes
    -------
    position is an array_like object, with shape (3,) or (n1, n2, .., 3).
    In all cases, the last dimension contains cordinates (x, y, z), in meter and in the lab frame
    """
    # - compute rotation matrix
    rotmat = rotation_matrix(u, theta)
    # - perform rotation
    rotated_position = np.tensordot(rotmat, position.T, axes=(1, 0)).T
    return rotated_position


def rotate_position_vector_alt(
    position: np.ndarray, u: np.ndarray, theta: float
) -> np.ndarray:
    """Rotates a position vector

    Alternative version to ``rotate_position_vector``, not using the
    numpy tensordot method, and therefore less performant. Included
    as a more transparent comparison, to make sure that tensordot is
    doing what we want.

    Parameters
    ----------
    position : array of shape (3,) or (n1, n2, ..., 3)
        cartesian coordinates in the lab frame
    u : array, shape (3,)
        the axis around which to perform the rotation
        it does not need to be normalized, the function will take
        care of it
    theta : float
        the angle for the rotation

    Returns
    -------
    rotated_position : array of shape (3,) or (n1, n2, ..., 3)
        rotated position vector

    Notes
    -------
    position is an array_like object, with shape (3,) or (n1, n2, .., 3).
    In all cases, the last dimension contains cordinates (x, y, z), in meter and in the lab frame
    """
    # - compute rotation matrix
    rotmat = rotation_matrix(u, theta)
    # - get coordinates
    X, Y, Z = position.T
    # - apply rotation matrix "by hand"
    Xrot = rotmat[0, 0] * X + rotmat[0, 1] * Y + rotmat[0, 2] * Z
    Yrot = rotmat[1, 0] * X + rotmat[1, 1] * Y + rotmat[1, 2] * Z
    Zrot = rotmat[2, 0] * X + rotmat[2, 1] * Y + rotmat[2, 2] * Z
    # - generate rotated position
    rotated_position = np.array([Xrot, Yrot, Zrot]).T
    return rotated_position


# % WRAPPERS


# -- ROTATION


def _rotation_get_value_decorator(get_value, u: np.ndarray, theta: float, vector: bool):
    """decorator for the get_value method for a rotation modifier

    Parameters
    ----------
    get_value : function
        the method to decorate
    u : array, shape (3,)
        the axis around which to perform the rotation
        it does not need to be normalized, the function will take
        care of it
    theta : float
        the angle for the rotation
        care of it
    vector : bool
        whether the result of 'get_value' is a vector
    """

    @wraps(get_value)
    def wrapped(position, nocheck=False, *args, **kwargs):
        position = check_position_array(position, nocheck)
        # note : we rotate the *coordinates* by -theta
        # to have the *object* rotated by +theta
        rotated_position = rotate_position_vector(position, u, -theta)
        value = get_value(rotated_position, *args, **kwargs)
        if vector:
            value = rotate_position_vector(value, u, theta)
        return value

    return wrapped


def _rotation_gen_infostring_obj_dectorator(
    gen_infostring_obj, u: np.ndarray, theta: float
):
    """decorator for the gen_infostring_obj method for a roation modifier
    see _rotation_get_value_decorator for the arguments definition
    """

    @wraps(gen_infostring_obj)
    def wrapped(*args, **kwargs):
        info = gen_infostring_obj(*args, **kwargs)
        info.add_element("<modifier> : rotation")
        info.add_element("<axis>", u)
        info.add_element("<theta>", theta)
        return info

    return wrapped


def rotate(obj: EnvObject, u: np.ndarray, theta: float):
    """Modifier for environment objects, performing a rotation

    Parameters
    ----------
    obj : EnvObject
        the object to rotate
    u : array, shape (3,)
        the axis around which to perform the rotation
        it does not need to be normalized, the function will take
        care of it
    theta : float
        the angle for the rotation

    Returns
    -------
    rotated_obj : EnvObject
        the rotated object
    """
    # - check
    if not isinstance(obj, EnvObject):
        raise TypeError("`obj` should be a EnvObject")

    # - decorate
    obj.get_value = _rotation_get_value_decorator(
        obj.get_value, u, theta, vector=obj.vector
    )
    obj.gen_infostring_obj = _rotation_gen_infostring_obj_dectorator(
        obj.gen_infostring_obj, u, theta
    )

    return obj


# -- SHIFT


def _shift_get_value_decorator(get_value, dr: np.ndarray):
    """decorator for the get_value method for a shift modifier

    Parameters
    ----------
    get_value : function
        the method to decorate
    dr : array, shape (3,)
        the value of the shift (cartesian coordinates)
    """

    @wraps(get_value)
    def wrapped(position, nocheck=False, *args, **kwargs):
        position = check_position_array(position, nocheck)
        # note : we shift the *coordinates* by -dr
        # to have the *object* shifted by +dr
        X, Y, Z = position.T
        dx, dy, dz = dr
        Xs = X - dx
        Ys = Y - dy
        Zs = Z - dz
        shifted_position = np.array([Xs, Ys, Zs]).T
        value = get_value(shifted_position, *args, **kwargs)
        return value

    return wrapped


def _shift_gen_infostring_obj_dectorator(gen_infostring_obj, dr: np.ndarray):
    """decorator for the gen_infostring_obj method for a shift modifier
    see _rotation_get_value_decorator for the arguments definition
    """

    @wraps(gen_infostring_obj)
    def wrapped(*args, **kwargs):
        info = gen_infostring_obj(*args, **kwargs)
        info.add_element("<modifier> : shift")
        info.add_element("<shift>", dr)
        return info

    return wrapped


def shift(obj: EnvObject, dr: np.ndarray):
    """Modifier for environment objects, performing a spatial shift

    Parameters
    ----------
    obj : EnvObject
        the object to rotate
    dr : array, shape (3,)
        the value of the shift (cartesian coordinates)

    Returns
    -------
    shifted_obj : EnvObject
        the shifted object
    """
    # - check
    if not isinstance(obj, EnvObject):
        raise TypeError("`obj` should be a EnvObject")

    # - decorate
    obj.get_value = _shift_get_value_decorator(obj.get_value, dr)
    obj.gen_infostring_obj = _shift_gen_infostring_obj_dectorator(
        obj.gen_infostring_obj, dr
    )

    return obj



"""wisdom
==================

as wise as one can be

"""

w = "ᄐᆊᄟᄍᅻ‘ᄂᄧᅍᅈჳᅐᅇᄴᅑᄃᅄᅓᅃᅁᅙᅘᅀჾᅃᄷᅛᅄᅁᅒᅃᄺᅊᄃᅊᄿᅕჳᅙᅋᄸჾᅈᅅᅎᅈᅁᅂᅕჳᅜᅈჳᅋᅃᄷᅊᄃᄴᅊᅑᅁᅌᄃᅇᅆᅇჳᅜᅄᅌჾ”ᅻᄏᄝᆊᄌ"
i = 4242
s = [80, 65, 83, 81, 65, 76]
e = print(
    "".join([chr(o) for o in [ord(l) - i - s[_ % len(s)] for _, l in enumerate(w)]])
)



"""The ``atomsmltr.environment`` subpackage provides classes to define the atom environment (lasers, mag. fields, zones).

Content
------------------

| ``atomsmltr.environment.fields``  : vector fields (magnetic fields & forces)
| ``atomsmltr.environment.lasers``  : laser beams
| ``atomsmltr.environment.zones``   : zones in position or speed space
"""

from .fields.magnetic import *
from .fields.force import *
from .lasers import *
from .zones import *



"""The ``atomsmltr.examples`` subpackage provides a collection of example configurations
for illustration & testing purposes

Examples
---------

Import the configuration from Feng et al 2024 Quantum Sci. Technol. 9 025017

.. code-block:: python

    from atomsmltr.examples.feng2024 import config
    from atomsmltr.examples.feng2024 import description
    print(description)

"""



"""Strontium
=============
Implements a dedicated class for Strontium atoms

>>> from atomsmltr.atoms import Strontium
"""

# % IMPORTS
import scipy.constants as csts

# % LOCAL IMPORTS
from ...atoms.generic import Atom
from ...atoms.transitions import J0J1Transition

# % CONSTANTS

STRONTIUM_88_MASS = 88 * csts.m_u  # kg
"""float: Strontium 88 mass (kg)"""

# - main transition
STRONTIUM_MAIN_WAVELENGTH = 460.862e-9  # m
"""float: Strontium main transition wavelength (m)"""
STRONTIUM_MAIN_GAMMA = 2 * csts.pi * 32.0e6  # rad/s
"""float: Strontium main transition natural linewidth (rad/s)"""
STRONTIUM_MAIN_LANDE_FACTOR = 1.0
"""float: Strontium main transition Lande factor"""

# - intercombination transition
STRONTIUM_INTERCOMBINATION_WAVELENGTH = 689.449e-9  # m
"""float: Strontium intercombination transition wavelength (m)"""
STRONTIUM_INTERCOMBINATION_GAMMA = 2 * csts.pi * 7.46e3  # rad/s
"""float: Strontium intercombination transition natural linewidth (rad/s)"""
STRONTIUM_INTERCOMBINATION_LANDE_FACTOR = 1.5
"""float: Strontium intercombination transition Lande factor"""

# %% TRANSITIONS


class MainLine(J0J1Transition):
    """The main (461nm) transition of Strontium"""

    def __init__(self):
        super().__init__(
            lande_factor=STRONTIUM_MAIN_LANDE_FACTOR,
            wavelength=STRONTIUM_MAIN_WAVELENGTH,
            Gamma=STRONTIUM_MAIN_GAMMA,
            tag="main",
        )


class IntercombinationLine(J0J1Transition):
    """The intercombination (689nm) transition of Strontium"""

    def __init__(self):
        super().__init__(
            lande_factor=STRONTIUM_INTERCOMBINATION_LANDE_FACTOR,
            wavelength=STRONTIUM_INTERCOMBINATION_WAVELENGTH,
            Gamma=STRONTIUM_INTERCOMBINATION_GAMMA,
            tag="intercombination",
        )


# %% ATOM


class Strontium(Atom):
    """Strontium 88 atomic class"""

    def __init__(self):

        # init super class
        super().__init__(
            mass=STRONTIUM_88_MASS,
            name="Strontium",
        )
        # add transitions
        main = MainLine()
        intercomb = IntercombinationLine()
        self.add_transition(main)
        self.add_transition(intercomb)



"""
forces
=======================

This module implements Forces field classes, which are mostly bare implementations of
generic fields defined in ``atomsmltr.environment.fields``.

See also
---------
atomsmltr.environment.fields.generic
"""

# % LOCAL IMPORTS
from ..generic import (
    Field,
    GradientField,
    ConstantField,
)

# % CLASSES

# -- FORCE PARENT CLASS
#   not really used currently, but will be useful if we need to
#   implement features specific to forces fields.
# > will also allow to check that the field is indeed a force
#   in the configuration object


class Force(Field):
    """A generic force field class. Used to set some properties common
    to all forces objects, and to have a way to identify magnetic field objects."""

    @property
    def type(self):
        return "Force"

    @property
    def unit(self):
        return "N"


# -- PERFECT FIELDS CLASSES
class ConstantForce(Force, ConstantField):
    """A constant Force

    See also
    ---------
    atomsmltr.environment.fields.generic.ConstantField
    """

    @property
    def type(self):
        return "constant force"


class GradientForce(Force, GradientField):
    """A perfect force gradient

    See also
    ---------
    atomsmltr.environment.fields.generic.GradientField
    """

    @property
    def type(self):
        return "gradient force"



"""Rubidium
=============
Implements a dedicated class for Rubidium atoms

>>> from atomsmltr.atoms import Rubidium
"""

# % IMPORTS
import scipy.constants as csts

# % LOCAL IMPORTS
from ...atoms.generic import Atom
from ...atoms.transitions import J0J1Transition

# % CONSTANTS

RUBIDIUM_87_MASS = 87 * csts.m_u  # kg
"""float: Rubidium 87 mass (kg)"""


# - D2 transition
RUBIDIUM_D2_WAVELENGTH = 780.241e-9  # m
"""float: Rubidium D2 transition wavelength (m)"""
RUBIDIUM_D2_GAMMA = 38.11e6  # rad/s
"""float: Rubidium D2 transition natural linewidth (rad/s)"""
RUBIDIUM_D2_LANDE_FACTOR = 1.0
"""float: Rubidium D2 transition Lande factor"""

# %% TRANSITIONS


class MainLine(J0J1Transition):
    """The D2 (780nm) transition of Rubidium"""

    def __init__(self):
        super().__init__(
            lande_factor=RUBIDIUM_D2_LANDE_FACTOR,
            wavelength=RUBIDIUM_D2_WAVELENGTH,
            Gamma=RUBIDIUM_D2_GAMMA,
            tag="main",
        )


# %% ATOM


class Rubidium(Atom):
    """Rubidium 87 atomic class"""

    def __init__(self):

        # init super class
        super().__init__(
            mass=RUBIDIUM_87_MASS,
            name="Rubidium",
        )
        # add transitions
        D2 = MainLine()
        self.add_transition(D2)



"""
interpolated fields
=======================

This module implements the generic ``InterpolatedField`` class

See also
---------
atomsmltr.environment.fields.magnetic
"""

# % IMPORTS
import numpy as np
from abc import abstractmethod
from scipy.interpolate import RegularGridInterpolator

# % LOCAL IMPORTS
from .generic import Field
from ...utils.infostring import InfoString

# % CLASS


class InterpolatedField(Field):
    """Creates an interpolated field from input values.

    Parameters
    ----------
    data_position : array
        the 'position' data of the field to interpolate
    data_field : array
        the 'value' data of the field to interpolate
    origin : array, shape (3,), optional
        cartesian coordinates of the origin for the field.
        If not set to (0,0,0), it will shift the interpolated field
        w.r.t the initial data, by default (0, 0, 0)
    scale : float, optional
        a scale factor for the interpolated field, by default 1.0
    tag : str, optional
        the field tag, by default None
    """

    def __init__(
        self,
        data_position: np.ndarray,
        data_field: np.ndarray,
        origin: np.ndarray = (0, 0, 0),
        scale: float = 1.0,
        tag: str = None,
    ):

        self.origin = origin
        self.scale = scale
        self.interpolate(data_position, data_field)
        super(InterpolatedField, self).__init__(tag)

    # -- requested method
    @abstractmethod
    def interpolate(self, data_position: np.ndarray, data_field: np.ndarray):
        """(re)initialize the interpolation

        Parameters
        ----------
        data_position : array
            the 'position' data of the field to interpolate
        data_field : array
            the 'value' data of the field to interpolate
        """
        pass

    @abstractmethod
    def _interp_fun(self, x: np.ndarray):
        """result of interpolation, has to be assigned via 'interpolate'"""
        pass

    # -- requested methods for Field
    # pylint : disable=method_hidden
    def _field_value_func(self, position):
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # most of the work is done by the '__interp_fun' method
        # here we only translate and scale
        translated_pos = position - self.origin
        value = self._interp_fun(translated_pos) * self.scale
        return value

    # -- getters and setters
    # -
    @property
    def origin(self) -> np.ndarray:
        """array, shape (,3): the origin for the interpolated field"""
        return self.__origin

    @origin.setter
    def origin(self, value: np.ndarray):
        self.__origin = self._check_3D_vector(value, "origin")

    # -
    @property
    def scale(self) -> float:
        """float: a scale factor for the field"""
        return self.__scale

    @scale.setter
    def scale(self, value: float):
        self.__scale = self._check_real_number(value, "scale")


class InterpolatedField1D1D(InterpolatedField):
    """Creates an interpolated 1D field from 1D input values.

    Parameters
    ----------
    data_position : array, shape (,n)
        the 'position' data of the field to interpolate
    data_field : array, shape (,n)
        the amplitude of the field to interpolate
    field_direction : array, shape (,3)
        the direction to which the interpolated field will point at
    position_direction : array, shape (,3)
        the direction corresponding to the data_position 'position' array
    origin : array, shape (3,), optional
        cartesian coordinates of the origin for the field.
        If not set to (0,0,0), it will shift the interpolated field
        w.r.t the initial data, by default (0, 0, 0)
    scale : float, optional
        a scale factor for the interpolated field, by default 1.0
    tag : str, optional
        the field tag, by default None
    """

    def __init__(
        self,
        data_position: np.ndarray,
        data_field: np.ndarray,
        field_direction: np.ndarray = (1, 0, 0),
        position_direction: np.ndarray = (1, 0, 0),
        origin: np.ndarray = (0, 0, 0),
        scale: float = 1.0,
        tag: str = None,
    ):
        self.field_direction = field_direction
        self.position_direction = position_direction

        super(InterpolatedField1D1D, self).__init__(
            data_position=data_position,
            data_field=data_field,
            origin=origin,
            scale=scale,
            tag=tag,
        )

    # -- Interp fun
    def _interp_fun(self, position: np.ndarray) -> np.ndarray:
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # - get X, Y, and Z
        x, y, z = position.T
        x, y, z = x.T, y.T, z.T

        # - get gradient vector angles
        theta = self.__xdir_theta
        phi = self.__xdir_phi

        # - get coordinates w.r.t origin
        x0, y0, z0 = self.origin
        xc = x - x0
        yc = y - y0
        zc = z - z0

        # compute coordinates in rotated frame
        # we want z
        z_rot = (
            xc * np.sin(theta) * np.cos(phi)
            + yc * np.sin(theta) * np.sin(phi)
            + zc * np.cos(theta)
        )
        value = self.__fun(z_rot)
        value = value[..., np.newaxis] * self.field_direction
        return value

    # -- GETTERS
    # -
    @property
    def data_position(self) -> np.ndarray:
        """array: x data used for interpolation"""
        return self.__data_position

    # -
    @property
    def data_field(self) -> np.ndarray:
        """array: y data used for interpolation"""
        return self.__data_field

    # -
    @property
    def field_direction(self) -> np.ndarray:
        """array, shape (3,): direction of the field"""
        return self.__field_direction

    @field_direction.setter
    def field_direction(self, value: np.ndarray):
        value = self._check_3D_vector(value, "field_direction", norm=True)
        self.__field_direction = value

    # -
    @property
    def position_direction(self) -> np.ndarray:
        """array, shape (,3) : the direction corresponding to the data_position 'position' array"""
        return self.__position_direction

    @position_direction.setter
    def position_direction(self, value: np.ndarray):
        value = self._check_3D_vector(value, "position_direction", norm=True)
        # compute angles
        ux, uy, uz = value
        theta = np.arctan2(np.sqrt(ux**2 + uy**2), uz)
        phi = np.arctan2(uy, ux)
        self.__xdir_theta = theta
        self.__xdir_phi = phi
        self.__position_direction = value

    # -- INTERPOLATE
    def interpolate(self, data_position, data_field):
        # -- check data
        # squeeze
        data_position = np.squeeze(data_position)
        data_field = np.squeeze(data_field)
        # dimension
        msg = "'data_position' and 'data_field' should be 1D arrays with same size"
        if data_position.ndim > 1 or data_field.ndim > 1:
            raise ValueError(msg)
        if data_position.shape != data_field.shape:
            raise ValueError(msg)
        # sort
        i_sort = np.argsort(data_position)
        data_position = data_position[i_sort]
        data_field = data_field[i_sort]

        # -- store
        self.__data_position = data_position
        self.__data_field = data_field

    def __fun(self, x):
        return np.interp(x, self.__data_position, self.__data_field)

    def gen_infostring_obj(self):
        """Generates an info string object"""
        unit = self.unit
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", "interpolated 1D-1D field")
        info.add_element("tag", self.tag)
        info.add_element("origin (m)", f"{self.origin}")
        info.add_element("field direction", f"{self.field_direction}")
        info.add_element("position direction", f"{self.position_direction}")
        info.add_element("scale", f"{self.scale:.3g}")
        info.add_element("position start", f"{self.data_position[0]:.3g}")
        info.add_element("position stop", f"{self.data_position[-1]:.3g}")

        return info


class InterpolatedField3D3D(InterpolatedField):
    """Creates an interpolated 3D field from 3D input values.

    Parameters
    ----------
    data_position : tuple of 3 ndarray of float, with shapes (m,), (n,) and (p,)
        the 'x', 'y' and 'z' position points of the field to interpolate
    data_field : array, shape (m,n,p,3)
        the magnetic field to interpolate
    origin : array, shape (3,), optional
        cartesian coordinates of the origin for the field.
        If not set to (0,0,0), it will shift the interpolated field
        w.r.t the initial data, by default (0, 0, 0)
    scale : float, optional
        a scale factor for the interpolated field, by default 1.0
    tag : str, optional
        the field tag, by default None
    """

    def __init__(
        self,
        data_position: tuple,
        data_field: np.ndarray,
        origin: np.ndarray = (0, 0, 0),
        scale: float = 1.0,
        tag: str = None,
    ):

        super(InterpolatedField3D3D, self).__init__(
            data_position=data_position,
            data_field=data_field,
            origin=origin,
            scale=scale,
            tag=tag,
        )

    # -- Interp fun
    def _interp_fun(self, position: np.ndarray) -> np.ndarray:
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        value = self.__fun(position)
        if position.ndim == 1:
            value = np.squeeze(value)
        return value

    # -- GETTERS
    # -
    @property
    def data_position(self) -> tuple:
        """tuple of 3 arrays: position data used for interpolation"""
        return self.__data_position

    # -
    @property
    def data_field(self) -> np.ndarray:
        """array: field data used for interpolation"""
        return self.__data_field

    # -- INTERPOLATE
    def interpolate(
        self,
        data_position: tuple,
        data_field: np.ndarray,
        fill_value: float = None,
        bounds_error: bool = False,
        **kwargs,
    ):
        """(re)initialize the interpolation

        Parameters
        ----------
        data_position : tuple of 3 ndarray of float, with shapes (m,), (n,) and (p,)
            the 'x', 'y' and 'z' position points of the field to interpolate
        data_field : array, shape (m,n,p,3)
            the magnetic field to interpolate
        fill_value : float, optional, defaults to None
            The value to use for points outside of the interpolation domain.
            If None, values outside the domain are extrapolated
        bounds_error : bool, optional, default to False
            If True, when interpolated values are requested outside of the domain of the input data, a ValueError is raised.
            If False, then fill_value is used.
        **kwargs :
            keyword arguments are passed to Scipy ``RegularGridInterpolator``
        """

        # -- check data
        # list
        if not isinstance(data_position, tuple):
            raise TypeError("'data_position' should be a tuple of three arrays")
        if len(data_position) != 3:
            raise ValueError("'data_position' should be a tuple of three arrays")

        # -- store
        self.__data_position = data_position
        self.__data_field = data_field

        # -- create interpolation object
        self.__fun = RegularGridInterpolator(
            data_position,
            data_field,
            fill_value=fill_value,
            bounds_error=bounds_error,
            **kwargs,
        )

    def gen_infostring_obj(self):
        """Generates an info string object"""
        unit = self.unit
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", "interpolated 3D-3D field")
        info.add_element("tag", self.tag)
        info.add_element("origin (m)", f"{self.origin}")
        info.add_element("scale", f"{self.scale:.3g}")
        info.add_element("data_field shape", f"{self.__data_field.shape}")
        return info



"""The ``atomsmltr.environment.fields`` subpackage provides definitions for vector fields

Currently, only magnetic fields are implemented, but this could be extented to electric fields

Examples
---------

Setup a magnetic field offset

.. code-block:: python

    from atomsmltr.environment.fields import MagneticOffset
    offset_field = MagneticOffset(field_value=(0,1,0), tag="offset")

See also
--------
atomsmltr.environment.fields.generic
atomsmltr.environment.fields.interpolated
atomsmltr.environment.fields.magnetic

"""

__all__ = [
    # - Magnetic Fields
    "MagneticGradient",
    "MagneticOffset",
    "MagneticQuadrupoleX",
    "MagneticQuadrupoleY",
    "MagneticQuadrupoleZ",
    "MagneticQuadrupole",
    "InterpMag1D1D",
    "InterpMag3D3D",
    # - Forces
    "Force",
    "GradientForce",
    "ConstantForce",
]

from .magnetic import (
    MagneticGradient,
    MagneticOffset,
    MagneticQuadrupoleX,
    MagneticQuadrupoleY,
    MagneticQuadrupoleZ,
    MagneticQuadrupole,
    InterpMag1D1D,
    InterpMag3D3D,
)

from .force import (
    Force,
    GradientForce,
    ConstantForce,
)



"""
transitions
=======================

This module implements the ``AtomicTransition`` class, that is meant to be embedded
in the ``Atom`` class.

Example
-------

.. code-block:: python

    from atomsmltr.atoms import Atom
    from atomsmltr.atoms.transitions import DummyTransition
    from scipy import constants as csts
    atom = Atom(mass=4 * csts.m_u, name="Helium")
    transition = DummyTransition("transition", Gamma=1, wavelength=1)
    atom.add_transition(transition, tag="transition")

See Also
--------
atomsmltr.atoms.generic.Atom
"""

# % IMPORTS
from abc import ABC, abstractmethod
import numpy as np
import scipy.constants as csts

# % LOCAL IMPORTS
from ..utils.infostring import InfoString

# % PHYSICS DEFINITIONS
"""In the following, we will define transitions using two parameters:
    > the wavelength in vacuum `lbda`
    > the natural linewidth `Gamma`

    All other parameters are derived from that
 """


def _w0(lbda: float) -> float:
    """returns the pulsation, in rad/s

    Parameters
    ----------
    lbda : float
        wavelength (m)

    Returns
    -------
    w0 : float
        pulsation (rad/s)
    """

    w0 = 2 * np.pi * csts.c / lbda
    return w0


def _Isat(lbda: float, Gamma: float) -> float:
    """Returns the saturation intensity, in W/m^2

    Parameters
    ----------
        lbda : float
            vacuum wavelength (in meters)
        Gamma : float
            natural linewidth (in rad/s)

    Returns
    -------
        Isat : float
            saturation intensity (in W/m^2)
    """
    w0 = _w0(lbda)
    Isat = csts.hbar * Gamma * w0**3 / 12 / np.pi / csts.c**2
    return Isat


def _Isat_mW_per_cm2(lbda: float, Gamma: float) -> float:
    """Returns the saturation intensity, in mW/cm^2

    Parameters
    ----------
        lbda : float
            vacuum wavelength (in meters)
        Gamma : float
            natural linewidth (in rad/s)

    Returns
    -------
        Isat: float
            saturation intensity (in mW/cm^2)
    """
    Isat_SI = _Isat(lbda, Gamma)
    Isat = Isat_SI * 1e3 / (1e2) ** 2
    return Isat


def _OmegaR(lbda: float, Gamma: float, I: float) -> float:
    """Returns the bare Rabi frequency for a two level system

    Parameters
    ----------
        lbda : float
            vacuum wavelength (in meters)
        Gamma : float
            natural linewidth (in rad/s)
        I : float
            saturation intensity (in W/m^2)

    Returns
    -------
        OmegaR : float
            the bare Rabi frequency (in rad/s)
    """
    Isat = _Isat(lbda, Gamma)
    OmegaR = Gamma * np.sqrt(I / 2 / Isat)
    return OmegaR


def _sat_param(lbda: float, Gamma: float, I: float, detuning: float) -> float:
    """Returns the saturation parameter for a two-level system.

    Beware, detuning is 2pi * (f_laser - f_transition)

    Parameters
    ----------
        lbda : float
            vacuum wavelength (in meters)
        Gamma : float
            natural linewidth (in rad/s)
        I : float
            saturation intensity (in W/m^2)
        detuning float
            laser detuning (in rad/s)

    Returns
    -------
        s : float
            the saturation parameter
    """
    Isat = _Isat(lbda, Gamma)
    s = (I / Isat) * (Gamma**2 / 4) / (detuning**2 + Gamma**2 / 4)
    return s


def _scattering_rate(lbda: float, Gamma: float, I: float, detuning: float) -> float:
    """Returns the scattering rate for a two-level system

    Beware, detuning is 2pi * (f_laser - f_transition)

    Parameters
    ----------
        lbda : float
            vacuum wavelength (in meters)
        Gamma : float
            natural linewidth (in rad/s)
        I : float
            saturation intensity (in W/m^2)
        detuning float
            laser detuning (in rad/s)

    Returns
    -------
        gamma_scatt : float
            the scattering rate (in /s)
    """
    s = _sat_param(lbda, Gamma, I, detuning)
    gamma_scatt = 0.5 * Gamma * s / (1 + s)
    return gamma_scatt


def _Doppler_temperature(Gamma: float, delta: float) -> float:
    """Returns the Doppler temperature for a transition


    Parameters
    ----------
        Gamma : float
            natural linewidth (in rad/s)

        delta : float
            laser detuning (in Hz)

    Returns
    -------
        T_Doppler : float
            Doppler temperature in K
    """
    T_Dopp = csts.hbar / 2 / csts.k * (delta**2 + Gamma**2 / 4) / np.abs(delta)
    return T_Dopp


# % ABSTRACT CLASSES


class AtomicTransition(ABC):
    """A generic (abstract) class to define electronic transitions in atoms

    Parameters
    ----------
    wavelength : float
        the vacuum wavelength (in m)
    Gamma : float
        the transition natural linewidth (in rad/s)
    tag : str
        a tag identifying the transition

    Notes
    ------
    This is an abstract class that cannot be used in simulations. For that purpose,
    see actual implementations of transitions.

    See also
    --------
    DummyTransition
    J0J1Transition
    """

    def __init__(self, wavelength: float, Gamma: float, tag: str):
        self.__tag = tag
        self.__wavelength = wavelength
        self.__Gamma = Gamma
        super().__init__()

    # -- READ-ONLY PROPERTIES
    # only with getters, no setters

    @property
    def tag(self):
        """str: transition identifier"""
        return self.__tag

    @property
    def wavelength(self):
        """float: transition vacuum wavelength (m)"""
        return self.__wavelength

    @property
    def Gamma(self):
        """float: transition natural linewidth (rad/s)"""
        return self.__Gamma

    @property
    def Isat(self):
        """float: transition saturation intensity (W/m^2)"""
        return _Isat(self.wavelength, self.Gamma)

    @property
    def Isat_mW_per_cm2(self):
        """float: transition saturation intensity (mW/cm^2)"""
        return _Isat_mW_per_cm2(self.wavelength, self.Gamma)

    @property
    def k(self):
        """float: transition wavenumber k = 2π / λ (m^-1)"""
        return 2 * np.pi / self.wavelength

    @property
    def Doppler_temperature(self):
        """float: Doppler temperature TD = hbar k / 2 / kB (K)"""
        return _Doppler_temperature(self.Gamma, -0.5 * self.Gamma)

    # -- METHODS

    def get_Doppler_temperature(self, detuning: float) -> float:
        """Returns the Doppler temperature for a given laser detuning

        Parameters
        ----------
        detuning : float
            laser detuning, in Hz

        Returns
        -------
        float
            the Doppler temperature, in K
        """
        return _Doppler_temperature(self.Gamma, detuning)

    def get_saturation_parameter(self, intensity: float, detuning: float) -> float:
        """Returns the saturation parameter (for a two-level system)

        Parameters
        ----------
            intensity : float
                laser intensity in W/m^2

            detuning : float
                laser detuning in rad/s

        Returns
        -------
            s : float
                the saturation parameter
        """
        s = _sat_param(self.wavelength, self.Gamma, intensity, detuning)
        return s

    @abstractmethod
    def get_scattering_rate(
        self,
        intensity: float,
        mag_field: float,
        polarization: np.ndarray,
        detuning: float,
    ):
        """Returns the scattering rate for a given laser / mag. field configuration

        Parameters
        ----------
        intensity : float
            laser intensity (W/m^)
        mag_field : float
            (scalar) magnetic field amplitude (T)
        polarization : ndarray, shape (,3)
            projection of the laser polarization on (pi, sigma+, sigma-)
        detuning : float
            laser detuning (rad/s)

        Returns
        -------
        scattering rate : float
            the transition scattering rate
        """
        pass

    @abstractmethod
    def get_resonant_speed(
        self,
        mag_field: float,
        polarization: str,
        detuning: float,
    ):
        """Returns the resonant speed for a given mag. field configuration

        Parameters
        ----------
        mag_field : float
            (scalar) magnetic field amplitude (T)
        polarization : str
            laser polarization : "pi", "sigma+" or "sigma-"
        detuning : float
            laser detuning (rad/s)

        Returns
        -------
        speed : float
            resonant speed (m/s)
        """
        pass

    def _gen_infostring_obj(self):
        """Generates an info string object"""
        info = InfoString(title=self.tag)
        info.add_section("Parameters")
        info.add_element("λ", f"{self.wavelength * 1e9:.2f} nm")
        info.add_element("Γ", f"2π × {self.Gamma / 2 / np.pi:.2e} Hz")
        info.add_element("Isat", f"{self.Isat_mW_per_cm2:.2f} mw/cm²")
        info.add_element("Doppler temp.", f"{self.Doppler_temperature:.2e} K")
        return info

    def gen_infostring_obj(self):
        """generates an ``InfoString`` object.

        Returns
        -------
        InfoString
            an ``InfoString`` object

        See also
        --------
        atomsmltr.utils.infostring.InfoString
        """
        return self._gen_infostring_obj()

    def gen_info_string(self, **kwargs):
        """generates an info string

        Returns
        -------
        info_string: str
            a string with information on the atom
        """
        return self.gen_infostring_obj().generate(**kwargs)

    def print_info(self):
        """prints the atom infostring"""
        print(self.gen_info_string())


class DummyTransition(AtomicTransition):
    """Dummy class, only for testing purposes"""

    def get_scattering_rate(
        self, intensity: float, mag_field: float, polarization: str, detuning: float
    ):
        """Returns the scattering rate for the DummyTransition model

        This is just a two-level atom with no dependence on mag. field or polarization

        Parameters
        ----------
        intensity : float
            laser intensity (W/m^)
        mag_field : float
            (scalar) magnetic field amplitude (T)
        polarization : ndarray, shape (,3)
            projection of the laser polarization on (pi, sigma+, sigma-)
        detuning : float
            laser detuning (rad/s)

        Returns
        -------
        scattering rate : float
            the transition scattering rate
        """
        rate = _scattering_rate(self.__wavelength, self.__Gamma, intensity, detuning)
        return rate

    def get_resonant_speed(
        self,
        mag_field: float,  # the amplitude of the magnetic field
        polarization: str,  # "pi", "sigma+", "sigma-"
        detuning: float,  # laser detuning
    ):
        """Returns the resonant speed for the DummyTransition model

        This actually always return zero...

        Parameters
        ----------
        mag_field : float
            (scalar) magnetic field amplitude (T)
        polarization : str
            laser polarization : "pi", "sigma+" or "sigma-"
        detuning : float
            laser detuning (rad/s)

        Returns
        -------
        speed : float
            resonant speed (m/s)
        """
        return 0


# % REAL IMPLEMENTATIONS


class J0J1Transition(AtomicTransition):
    """A class to handle J=0 -> J=1 transitions

    Parameters
    ----------
    wavelength : float
        the vacuum wavelength (in m)
    Gamma : float
        the transition natural linewidth (in rad/s)
    lande_factor : float
        the lande g-factor for the transition
    tag : str
        a tag identifying the transition
    """

    def __init__(self, wavelength: float, Gamma: float, lande_factor: float, tag: str):
        self.__lande_factor = lande_factor
        super().__init__(wavelength=wavelength, Gamma=Gamma, tag=tag)

    @property
    def lande_factor(self):
        """float: the lande g-factor for the transition"""
        return self.__lande_factor

    def gen_infostring_obj(self):
        info = self._gen_infostring_obj()
        info.add_element("lande factor g", f"{self.lande_factor}")
        return info

    def get_scattering_rate(
        self,
        intensity: float,  # the intensity in W/cm^2
        mag_field: float,  # the amplitude of the magnetic field
        polarization: list,  # projection (squared) of laser polarization on (pi, sigma+, sigma-)
        detuning: float,  # laser detuning (in rad/s !!!!!!)
    ):
        # -- get projections
        # TODO : checks here
        polarization = np.asanyarray(polarization)
        proj_pi, proj_sigm_plus, proj_sigm_minus = polarization.T
        proj_pi = proj_pi.T
        proj_sigm_minus = proj_sigm_minus.T
        proj_sigm_plus = proj_sigm_plus.T

        # -- Zeeman effect
        # NB : detuning is 2 * pi * (f_laser - f_atom)
        # constants
        mu_B = csts.physical_constants["Bohr magneton"][0]
        mu = self.lande_factor * mu_B / csts.hbar

        # compute detuning
        det_pi = detuning
        det_sigm_minus = detuning + mu * mag_field
        det_sigm_plus = detuning - mu * mag_field

        # -- Compute scattering rate
        # NB : we assume that the transition is not saturated and we can sum
        # all the polarization components
        scatt_pi = _scattering_rate(
            self.wavelength, self.Gamma, intensity * proj_pi, det_pi
        )
        scatt_sigm_minus = _scattering_rate(
            self.wavelength, self.Gamma, intensity * proj_sigm_minus, det_sigm_minus
        )
        scatt_sigm_plus = _scattering_rate(
            self.wavelength, self.Gamma, intensity * proj_sigm_plus, det_sigm_plus
        )

        # sum
        scatt_total = scatt_pi + scatt_sigm_minus + scatt_sigm_plus

        return scatt_total

    def get_resonant_speed(
        self,
        mag_field: float,  # the amplitude of the magnetic field
        polarization: str,  # "pi", "sigma+", "sigma-"
        detuning: float,  # laser detuning
    ):
        # -- check input
        polar_list = ["pi", "sigma+", "sigma-"]
        msg = f"'polarization' should be in {polar_list}"
        assert polarization in polar_list, msg

        # -- factor
        mu_B = csts.physical_constants["Bohr magneton"][0]
        mu = self.lande_factor * mu_B / csts.hbar
        prefact = {"pi": 0, "sigma+": 1, "sigma-": -1}

        v_res = (detuning - mu * prefact[polarization] * mag_field) / self.k
        return v_res



"""The ``atomsmltr.environment.fields.magnetic`` subpackage provides definitions for magnetic fields

Those are mostly direct implementations of generic `Fields` objects
from `atomsmltr.environment.fields.generic`

Examples
---------

Setup a magnetic field offset

.. code-block:: python

    from atomsmltr.environment.fields import MagneticOffset
    offset_field = MagneticOffset(offset=(0,1,0), tag="offset")

See also
--------
atomsmltr.environment.fields.generic
atomsmltr.environment.fields.interpolated

"""

__all__ = [
    "MagneticField",
    "MagneticGradient",
    "MagneticOffset",
    "MagneticQuadrupoleX",
    "MagneticQuadrupoleY",
    "MagneticQuadrupoleZ",
    "MagneticQuadrupole",
    "InterpMag1D1D",
    "InterpMag3D3D",
]
from .magnetic import (
    MagneticField,
    MagneticGradient,
    MagneticOffset,
    InterpMag1D1D,
    InterpMag3D3D,
    MagneticQuadrupoleX,
    MagneticQuadrupoleY,
    MagneticQuadrupoleZ,
    MagneticQuadrupole,
)



"""Home-made deterministic integrators
=========================================

Implements homemade integrators for deterministic systems, that is, not taking into
acount diffusion due to photon scattering.
"""

# % IMPORTS
import numpy as np
from functools import partial
from abc import abstractmethod

# % LOCAL IMPORTS
from .simbase import Simulation, SimRes, get_force_vec
from ..configurator import Configuration


# % HOME-MADE SIMULATORS


def stop_position_event(u: np.ndarray, stop_position: list):
    """Implements 'stop' events for home-made simulators, based on atom's position

    Parameters
    ----------
    u : array, shape (n,m,...,6)
        position/speed vector, according to our vectorization convention
    stop_position : list
        list of Zones objects targetting position with actions set to stop

    Returns
    -------
    res: bool
        whether to stop the simulation

    See also
    --------
    atomsmltr.environment.zones
    atomsmltr.simulation.configurator.Configuration.get_stop_zones()
    """
    x, y, z, _, _, _ = u.T
    position = np.array([x, y, z]).T
    res = np.logical_and.reduce([zone.get_value(position) for zone in stop_position])
    res = res
    return res


def stop_speed_event(u: np.ndarray, stop_speed: list):
    """Implements 'stop' events for home-made simulators, based on atom's speed

    Parameters
    ----------
    u : array, shape (n,m,...,6)
        position/speed vector, according to our vectorization convention
    stop_speed : list
        list of Zones objects targetting speed with actions set to stop

    Returns
    -------
    res: bool
        whether to stop the simulation

    See also
    --------
    atomsmltr.environment.zones
    atomsmltr.simulation.configurator.Configuration.get_stop_zones()
    """
    _, _, _, vx, vy, vz = u.T
    speed = np.array([vx, vy, vz]).T
    res = np.logical_and.reduce([zone.get_value(speed) for zone in stop_speed])
    res = res
    return res


class CustomSimulationBase(Simulation):
    """A Base for home-made deterministic simulations

    Not meant to be used directly, just gathering common method
    """

    def __init__(
        self,
        config: Configuration = None,
    ):
        super(CustomSimulationBase, self).__init__(config)

    def get_force(self, u):
        force = get_force_vec(u, self.config)
        return force

    def dudt(self, t, u):
        F = self.get_force(u)
        _, _, _, vx, vy, vz = u.T
        dx, dy, dz = vx, vy, vz
        dvx, dvy, dvz = F.T / self.config.atom.mass
        res = np.array([dx, dy, dz, dvx, dvy, dvz]).T
        return res

    @abstractmethod
    def _iterate(self, t, u, dt):
        """returns the evolution du of u between t and t+dt"""

    def _integrate(self, u0, t):
        # - u0 to array
        u = np.asanyarray(u0)
        # - get stop events
        events = []
        stop_position, stop_speed = self.config.get_stop_zones()
        if stop_position:
            stop_pos = partial(stop_position_event, stop_position=stop_position)
            events.append(stop_pos)
        if stop_speed:
            stop_sp = partial(stop_speed_event, stop_speed=stop_speed)
            events.append(stop_sp)
        # - time
        # TODO : add checks on time
        t = np.asanyarray(t)
        t = np.sort(t)
        dt = np.diff(t)
        # - initialize
        y = np.empty((*u.shape, len(t)))
        y[..., 0] = u
        stop = False
        # - integrate
        i = 1
        u_none = np.full((6,), np.nan)
        for i, (tt, h) in enumerate(zip(t[1:], dt)):
            # check events
            if events:
                for ev in events:
                    test = ev(u)
                    u[np.logical_not(test), :] = u_none
                    if not np.any(test):
                        stop = True
            if stop:
                break

            # perform step
            u = u + self._iterate(tt, u, h)
            y[..., i + 1] = u

        if stop:
            y = y[..., : i + 1]
            t = t[: i + 1]

        res = SimRes(t=t, y=y)

        return res

    def _u0_list_checker(self, value):
        if not hasattr(value, "__iter__"):
            raise ValueError("'u0_list' should be an iterable object")
        if value:
            for u0 in value:
                if np.asanyarray(u0).shape != (6,):
                    raise ValueError("'u0_list' should be a list of arrays of size 6")
        return value


class RK4(CustomSimulationBase):
    """A homemade simulator based on fourth order Runge-Kutta method

    Parameters
    ----------
    config : Configuration, optional
        the configuration to consider for the simulation

    References
    ----------
    https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods

    """

    def __init__(
        self,
        config: Configuration = None,
    ):
        super(RK4, self).__init__(config)

    def _iterate(self, t, u, dt):
        """returns the evolution du of u between t and t+dt
        Here we use the fourth order Runge-Kutta method
        """
        # perform step
        k1 = self.dudt(t, u)
        k2 = self.dudt(t + 0.5 * dt, u + 0.5 * k1 * dt)
        k3 = self.dudt(t + 0.5 * dt, u + 0.5 * k2 * dt)
        k4 = self.dudt(t + dt, u + k3 * dt)
        du = (dt / 6) * (k1 + 2 * k2 + 2 * k3 + k4)
        return du


class Euler(CustomSimulationBase):
    """A homemade simulator based on Euler's method

    Parameters
    ----------
    config : Configuration, optional
        the configuration to consider for the simulation

    References
    ----------
    TODO: put here

    """

    def __init__(
        self,
        config: Configuration = None,
    ):
        super(Euler, self).__init__(config)

    def _iterate(self, t, u, dt):
        """returns the evolution du of u between t and t+dt
        Here we use the Euler method
        """
        F = self.get_force(u)
        _, _, _, vx, vy, vz = u.T
        dvx, dvy, dvz = F.T / self.config.atom.mass * dt
        dx = vx * dt + 0.5 * dvx * dt
        dy = vy * dt + 0.5 * dvy * dt
        dz = vz * dt + 0.5 * dvz * dt
        du = np.array([dx, dy, dz, dvx, dvy, dvz]).T
        return du


class VelocityVerlet(CustomSimulationBase):
    """A homemade simulator based on velocity verlet method

    Parameters
    ----------
    config : Configuration, optional
        the configuration to consider for the simulation

    References
    ----------
    TODO: put here

    """

    def __init__(
        self,
        config: Configuration = None,
    ):
        super(VelocityVerlet, self).__init__(config)

    def _iterate(self, t, u, dt):
        """returns the evolution du of u between t and t+dt
        Here we use the Velocity verlet method
        """
        # 1) compute next position
        F = self.get_force(u)
        x, y, z, vx, vy, vz = u.T
        ax, ay, az = F.T / self.config.atom.mass
        dx = vx * dt + 0.5 * ax * dt**2
        dy = vy * dt + 0.5 * ay * dt**2
        dz = vz * dt + 0.5 * az * dt**2
        # 2) compute next speed
        u_next = np.array([x + dx, y + dy, z + dz, vx, vy, vz]).T
        F_next = self.get_force(u_next)
        ax_next, ay_next, az_next = F_next.T / self.config.atom.mass
        dvx = 0.5 * (ax + ax_next) * dt
        dvy = 0.5 * (ay + ay_next) * dt
        dvz = 0.5 * (az + az_next) * dt
        du = np.array([dx, dy, dz, dvx, dvy, dvz]).T
        return du



"""The ``atomsmltr.utils`` subpackage provides several tools that are useful
accross the whole module

See also
---------
atomsmltr.utils.infostring
atomsmltr.utils.misc
"""

__all__ = [
    "InfoString",
    "Plottable",
]

from .infostring import InfoString
from .plotter import Plottable



"""
Examples : Chen 2021
=======================

This file provides the configuration for the 1D MOT, 1D molasses and 3D MOT configurations
in Chen 2021  : arXiv:2105.06447

"""

# % EXPERIENCE DESCRIPTION
description = """

This code includes three example simulations from the AtomECS paper (Chen et al., arXiv:2105.06447), demonstrating
atomic motion under different laser cooling configurations.

1 ) One-dimensionnal molasses :

>> from atomsmltr.examples.chen2021 import config_1D_molasses

This simulation models a one-dimensional optical molasses for 87Rb atoms, following the setup
described in the AtomECS paper. The system consists of two counter-propagating, red-detuned laser beams
along the z-axis, forming an optical molasses that slows atomic motion through Doppler cooling.

Physical configuration:
- Atom species : 87Rb (D2 cooling transition at 780 nm)
- Laser power : 10 mW per beam
- Beam waist (1/e radius) : 1 cm
- Detuning : -12 MHz from resonance
- Polarization : σ-

2 ) One-dimensional Magneto-Optical Trap (1D MOT):

>> from atomsmltr.examples.chen2021 import config_1D_MOT

This simulation models a one-dimensional magneto-optical trap for 88Sr atoms, extending the
optical molasses setup by adding a magnetic quadrupole field along the z-axis. The combination
of red-detuned counter-propagating laser beams and the position-dependent magnetic field
allows atoms with velocities below the capture velocity to be trapped and cooled.

Physical configuration:
- Atom species : 88Sr (main cooling transition at 461 nm)
- Laser power : 30 mW per beam
- Beam waist (1/e radius): 1 cm
- Detuning : -12 MHz from resonance
- Polarization : σ-
- Magnetic field type : Perfect quadrupole
- Magnetic field direction : (0, 0, 1)
- Magnetic field origin : (0, 0, 0)
- Magnetic field gradient : 0.15 T/m


3 ) Three-dimensional MOT :

>> from atomsmltr.examples.chen2021 import config_3D_MOT

This simulation models a three-dimensional MOT for 87Rb atoms using six counter-propagating
laser beams along the Cartesian axes, combined with a 3D quadrupole magnetic field with strong axis along the z-axis.

Physical configuration:
- Atom species : 87Rb (D2 cooling transition at 461 nm)
- Laser power : 20 mW per beam
- Beam waist (1/e radius) : 1 cm
- Detuning : -12 MHz from resonance
- Polarization for lasers along the z-axis : σ-
- Polarization for lasers along the other axis : σ+
- Magnetic field type : Perfect quadrupole
- Magnetic field direction : (0, 0, 1)
- Magnetic field origin : (0, 0, 0)
- Magnetic field gradient : 0.3 T/m

"""

# % IMPORTS
import numpy as np
import matplotlib.pyplot as plt

# % LOCAL IMPORTS
from atomsmltr.environment import GaussianLaserBeam
from atomsmltr.atoms import Strontium, Rubidium
from atomsmltr.simulation import Configuration
from atomsmltr.environment.lasers import CircularLeft, CircularRight
from atomsmltr.environment.fields.magnetic import MagneticQuadrupoleZ

# --------------------------------------------------------------------------------------------------------

# % GENERATE CONFIGURATION of the 1D molasses


# -- init config with strontium atom
atom_rubidium = Rubidium()

# -- get Strontium main transition information
main = atom_rubidium.trans["main"]


# -- setup lasers of the 1D molasses
# cf. config from Chen 2021
laser_1_molasses = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=(1e-2) * np.sqrt(2),
    power=1e-2,
    waist_position=(0, 0, 0),
    direction=(0, 0, 1),
    polarization=CircularLeft(),
    tag="las1molasses",
)

laser_2_molasses = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=(1e-2) * np.sqrt(2),
    power=1e-2,
    waist_position=(0, 0, 0),
    direction=(0, 0, -1),
    polarization=CircularLeft(),
    tag="las2molasses",
)


# -- add everything to the configuration

config_1D_molasses = Configuration()
config_1D_molasses.atom = atom_rubidium

# add objects
config_1D_molasses += laser_1_molasses, laser_2_molasses

# setup atomlight
config_1D_molasses.add_atomlight_coupling("las1molasses", "main", -2 * np.pi * 6e6)
config_1D_molasses.add_atomlight_coupling("las2molasses", "main", -2 * np.pi * 6e6)


# --------------------------------------------------------------------------------------------------------

# % GENERATE CONFIGURATION of the 1D MOT


# -- init config with strontium atom
atom_strontium = Strontium()

# -- get Strontium main transition information
main = atom_strontium.trans["main"]

# -- setup magnetic field: perfect quadrupole with a strong Z-axis in Chen 2021
# Define magnet properties
origin_1D = np.array((0, 0, 0))
gradient_1D = 0.15  # T/m
mag_field_1D = MagneticQuadrupoleZ(
    origin=origin_1D, slope=gradient_1D, tag="mag_field_1D"
)

# -- setup lasers of the 1D MOT
# cf. config from Chen 2021
laser_1_MOT = GaussianLaserBeam(
    wavelength=460.862e-9,
    waist=(1e-2) * np.sqrt(2),
    power=3e-2,
    waist_position=(0, 0, 0),
    direction=(0, 0, 1),
    polarization=CircularLeft(),
    tag="las1mot",
)

laser_2_MOT = GaussianLaserBeam(
    wavelength=460.862e-9,
    waist=(1e-2) * np.sqrt(2),
    power=3e-2,
    waist_position=(0, 0, 0),
    direction=(0, 0, -1),
    polarization=CircularLeft(),
    tag="las2mot",
)


# -- add everything to the configuration

config_1D_MOT = Configuration()
config_1D_MOT.atom = atom_strontium

# add objects
config_1D_MOT += laser_1_MOT, laser_2_MOT, mag_field_1D

# setup atomlight
config_1D_MOT.add_atomlight_coupling("las1mot", "main", -2 * np.pi * 12e6)
config_1D_MOT.add_atomlight_coupling("las2mot", "main", -2 * np.pi * 12e6)


# --------------------------------------------------------------------------------------------------------

# % GENERATE CONFIGURATION of the 3D MOT


# -- init config with strontium atom
atom_rubidium = Rubidium()

# -- get Strontium main transition information
main = atom_rubidium.trans["main"]

# -- setup magnetic field: perfect quadrupole with a strong Z-axis in Chen 2021
# Define magnet properties
origin_3D = np.array((0, 0, 0))
gradient_3D = 0.3  # T/m
mag_field_3D = MagneticQuadrupoleZ(
    origin=origin_3D, slope=gradient_3D, tag="mag_field_3D"
)

# -- setup lasers of the 1D MOT
# cf. config from Chen 2021
laser_1_3D_MOT = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=66.7e-3,
    power=0.02,
    waist_position=(0, 0, 0),
    direction=(0, 0, 1),
    polarization=CircularLeft(),
    tag="las1mot3D",
)

laser_2_3D_MOT = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=66.7e-3,
    power=0.02,
    waist_position=(0, 0, 0),
    direction=(0, 0, -1),
    polarization=CircularLeft(),
    tag="las2mot3D",
)

laser_3_3D_MOT = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=66.7e-3,
    power=0.02,
    waist_position=(0, 0, 0),
    direction=(1, 0, 0),
    polarization=CircularRight(),
    tag="las3mot3D",
)

laser_4_3D_MOT = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=66.7e-3,
    power=0.02,
    waist_position=(0, 0, 0),
    direction=(-1, 0, 0),
    polarization=CircularRight(),
    tag="las4mot3D",
)

laser_5_3D_MOT = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=66.7e-3,
    power=0.02,
    waist_position=(0, 0, 0),
    direction=(0, 1, 0),
    polarization=CircularRight(),
    tag="las5mot3D",
)
laser_6_3D_MOT = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=66.7e-3,
    power=0.02,
    waist_position=(0, 0, 0),
    direction=(0, -1, 0),
    polarization=CircularRight(),
    tag="las6mot3D",
)


# -- add everything to the configuration

config_3D_MOT = Configuration()
config_3D_MOT.atom = atom_rubidium

# add objects
config_3D_MOT += (
    laser_1_3D_MOT,
    laser_2_3D_MOT,
    laser_3_3D_MOT,
    laser_4_3D_MOT,
    laser_5_3D_MOT,
    laser_6_3D_MOT,
    mag_field_3D,
)

# setup atomlight
config_3D_MOT.add_atomlight_coupling("las1mot3D", "main", -2 * np.pi * 3e6)
config_3D_MOT.add_atomlight_coupling("las2mot3D", "main", -2 * np.pi * 3e6)
config_3D_MOT.add_atomlight_coupling("las3mot3D", "main", -2 * np.pi * 3e6)
config_3D_MOT.add_atomlight_coupling("las4mot3D", "main", -2 * np.pi * 3e6)
config_3D_MOT.add_atomlight_coupling("las5mot3D", "main", -2 * np.pi * 3e6)
config_3D_MOT.add_atomlight_coupling("las6mot3D", "main", -2 * np.pi * 3e6)



"""Polarization
==================

Here we implement the ``Polarization`` class that we use to define the polarization of lasers

Remarks
--------

We define the polarization in the **frame of the laser**, with laser propagation along z.

We denote **x** as the **'horizontal'** axis and **y** as the **'vertical'** axis.

For circular polarizations, we take the **observer convention**.

To have a combined formalism for all polarization, in the end we define a **polarization vector** ``p_vec``,
following the Poincarré formalism. **ATTENTION:** there might be different ways of defining this vector,
refer to the package documentation for a thorough definition.

In the case of 'linear' polarization, an additionnal argument ``angle`` has to be provided, that gives
the angle of the linear polarization with respect to the x axis. Hence ``angle = 0`` corresponds to a
linear polarization along x, and ``angle = pi/2`` to a linear polarization along y

In the case of 'vector' polarization, polarization vector has to be given with the ``vec`` argument. ``vec``
are the cartesian coordinates of the vector in the (x,y,z) basis. For instance :

| ``> vec = (1, 0, 0)``  : linear polarization along x
| ``> vec = (0, 1, 0)``  : linear polarization along y
| ``> vec = (0, 0, 1)``  : circular right polarization
| ``> vec = (0, 0, -1)`` : circular left polarization

**ATTENTION** : for ease of use, the vector does not have to be normalized, but the resulting one will
be.


"""

# % IMPORTS
import numpy as np
from abc import ABC, abstractmethod

# % LOCAL IMPORTS
from ...utils.infostring import InfoString


# % ABSTRACT CLASS


class Polarization(ABC):
    """An object to handle laser polarization."""

    def __init__(self):
        self._vector = None
        self._u = None
        self._v = None

    # -- PROPERTIES
    @property
    def vector(self):
        vector = np.asanyarray(self._vector)
        return vector

    # -- METHODS

    def get_polarization_vector_angles(self) -> tuple:
        """Returns the angles describing the current polarization vector.

        (see documentation for thorough description)

        Returns
        -------
        u, v : floats
            the u (polar) and v (azimuthal) angles

        Notes
        -----
        The polarization is decribed in the Poincarré/Bloch-like sphere as a vector.
        This function yields the angles u (polar) and v (azimuthal)

        Note that we do not use theta or phi as those angles are already used to
        describe the orientation of the laser propagation vector in the ``LaserBeam`` class

        """
        u = self._u
        v = self._v
        return u, v

    def refresh_polarization_vector_angles(self):
        """Updates the polarization vector angles u & v to match the current value of the
        polarization vector.
        """
        x, y, z = self.vector
        u = np.arctan2(np.sqrt(x**2 + y**2), z)
        v = np.arctan2(y, x)
        self._u = u
        self._v = v

    def get_polarization_vector_projection(self, target: str) -> complex:
        """Returns the scalar projection of the current polarization vector on a target polarization state

        Parameters
        ----------
        target : str
            the state on which to project (see docstring Notes)

        Returns
        -------
        proj: complex
            the projection

        Notes
        -----
        The polarization Psi is defined as :

            |Psi⟩ = exp(-i*v) cos(u/2) |R⟩ +  exp(i*v) sin(u/2) |L⟩

        with |R⟩, |L⟩ the right- and left-handed circular polarization states. We also have

            |x⟩ = |V⟩ = (1/sqrt(2)) (|L⟩ + |R⟩)
            |y⟩ = |H⟩ = (i/sqrt(2)) (|L⟩ - |L⟩)

        Target should refer to the special polarization states defined in the class :

        >>> 'vertical', 'horizontal', 'circular left', 'circular right'

        and corresponding shorthands:

        >>> 'V' or 'x', 'H' or 'y', 'R', 'L'

        """
        # get angle values
        u, v = self.get_polarization_vector_angles()
        # common calculations
        A = np.exp(-1j * v) * np.cos(u / 2)
        B = np.exp(1j * v) * np.sin(u / 2)
        # return projection on desired vector
        match target.upper():
            case "V" | "X" | "VERTICAL":
                proj = (A + B) / np.sqrt(2)
            case "H" | "Y" | "HORIZONTAL":
                proj = 1j * (A - B) / np.sqrt(2)
            case "R" | "CIRCULAR RIGHT":
                proj = A
            case "L" | "CIRCULAR LEFT":
                proj = B
            case _:
                GOOD = ["vertical", "horizontal", "circular left", "circular right"]
                raise ValueError(f"Wrong value for target state, shoud be in {GOOD}")

        return proj

    def get_polarization_vector_projection_norm(self, target: str) -> float:
        """Returns the squared norm of scalar projection of the current polarization vector on a target polarization state

        Parameters
        ----------
        target : str
            the state on which to project (see docstring Notes)

        Returns
        -------
        norm: float
            the squared norm of the projection

        Notes
        -----
        The polarization Psi is defined as :

            |Psi⟩ = exp(-i*v) cos(u/2) |R⟩ +  exp(i*v) sin(u/2) |L⟩

        with |R⟩, |L⟩ the right- and left-handed circular polarization states. We also have

            |x⟩ = |V⟩ = (1/sqrt(2)) (|L⟩ + |R⟩)
            |y⟩ = |H⟩ = (i/sqrt(2)) (|L⟩ - |L⟩)

        Target should refer to the special polarization states defined in the class :

        >>> 'vertical', 'horizontal', 'circular left', 'circular right'

        and corresponding shorthands:

        >>> 'V' or 'x', 'H' or 'y', 'R', 'L'

        """
        # get angle values
        u, v = self.get_polarization_vector_angles()
        # return projection on desired vector
        match target.upper():
            case "V" | "X" | "VERTICAL":
                norm = 0.5 * (1 + 2 * np.cos(u / 2) * np.sin(u / 2) * np.cos(2 * v))
            case "H" | "Y" | "HORIZONTAL":
                norm = 0.5 * (1 - 2 * np.cos(u / 2) * np.sin(u / 2) * np.cos(2 * v))
            case "R" | "CIRCULAR RIGHT":
                norm = np.cos(u / 2) ** 2
            case "L" | "CIRCULAR LEFT":
                norm = np.sin(u / 2) ** 2
            case _:
                GOOD = ["vertical", "horizontal", "circular left", "circular right"]
                raise ValueError(f"Wrong value for target state, shoud be in {GOOD}")

        return norm

    def gen_infostring_obj(self):
        """Returns an info string object for the current polarization state"""
        # - init InfoString object
        info = InfoString("POLARIZATION PROPERTIES")
        # - populate info string
        # object settings
        info.add_section("Polarization settings")
        if isinstance(self, Linear):
            info.add_element("type", self.type)
            info.add_element("angle", f"{self.angle / np.pi:.2f} pi")
        elif isinstance(self, Vector):
            info.add_element("type", self.type)
            info.add_element("vector", self.vector)
        else:
            info.add_element("type", self.type)

        # vector
        info.add_section("Polarization vector")
        u, v = self.get_polarization_vector_angles()
        x, y, z = self.vector
        info.add_element("coords", f"({x:.2f}, {y:.2f}, {z:.2f})")
        info.add_element("polar angle u", f"{u/np.pi:.2f} pi")
        info.add_element("azimt angle v", f"{v/np.pi:.2f} pi")

        # Projections (amplitudes)
        u, v = self.get_polarization_vector_angles()
        info.add_section("Projections (amplitudes)")
        for target in ["vertical", "horizontal", "circular left", "circular right"]:
            proj = self.get_polarization_vector_projection(target)
            if target == "circular right":
                info.add_element(target, f"{proj:.2f}")
            else:
                info.add_element(target, f"{proj:.2f}")

        # Projections (norm)
        u, v = self.get_polarization_vector_angles()
        info.add_section("Projections (squared norm)")
        for target in ["vertical", "horizontal", "circular left", "circular right"]:
            proj = self.get_polarization_vector_projection_norm(target)
            if target == "circular right":
                info.add_element(target, f"{proj:.2f}")
            else:
                info.add_element(target, f"{proj:.2f}")

        return info

    def gen_info_string(self, **kwargs):
        return self.gen_infostring_obj().generate(**kwargs)

    def print_info(self):
        print(self.gen_info_string())


# % ACTUAL IMPLEMENTATIONS


class Vertical(Polarization):
    """Vertical polarization (along x in the laser frame)"""

    def __init__(self):
        super(Vertical, self).__init__()
        self.type = "Vertical"
        self._vector = (1, 0, 0)
        self.refresh_polarization_vector_angles()


class Horizontal(Polarization):
    """Horizontal polarization (along y in the laser frame)"""

    def __init__(self):
        super(Horizontal, self).__init__()
        self.type = "Horizontal"
        self._vector = (0, 1, 0)
        self.refresh_polarization_vector_angles()


class CircularLeft(Polarization):
    """Circular Left polarization (observer point of vue)"""

    def __init__(self):
        super(CircularLeft, self).__init__()
        self.type = "Circular Left"
        self._vector = (0, 0, -1)
        self.refresh_polarization_vector_angles()


class CircularRight(Polarization):
    """Circular Right polarization (observer point of vue)"""

    def __init__(self):
        super(CircularRight, self).__init__()
        self.type = "Circular Right"
        self._vector = (0, 0, 1)
        self.refresh_polarization_vector_angles()


class Linear(Polarization):
    """Arbitrary linear polarization.

    Here, ``angle`` is the angle of the linear polarization with respect to the x axis.
    Hence ``angle = 0`` corresponds to a linear polarization along x, and ``angle = pi/2`` to a linear polarization along y

    Parameters
    ----------
    angle : float
        angle of the arbitrary linear polarization w.r.t the x axis (radians)
    """

    def __init__(self, angle: float):
        super(Linear, self).__init__()
        self.type = "Linear"
        self.angle = angle

    @property
    def angle(self) -> float:
        """float: angle of the arbitrary linear polarization w.r.t the x axis (radians)"""
        return self._angle

    @angle.setter
    def angle(self, value: float) -> None:
        # convert int into float
        if isinstance(value, int):
            value = float(value)

        if not isinstance(value, float):
            raise ValueError("Angle must be a float")

        self._angle = value
        self._vector = (np.cos(self.angle), np.sin(self.angle), 0)
        self.refresh_polarization_vector_angles()


class Vector(Polarization):
    """Allows to define an arbitrary polarization.

    The polarization vector has to be given with the ``vector`` argument.
    ``vector`` are the cartesian coordinates of the vector in the (x,y,z) basis.

    For instance :

    | ``> vec = (1, 0, 0)``  : linear polarization along x
    | ``> vec = (0, 1, 0)``  : linear polarization along y
    | ``> vec = (0, 0, 1)``  : circular right polarization
    | ``> vec = (0, 0, -1)`` : circular left polarization

    **ATTENTION** : for ease of use, the vector does not have to be normalized, but the resulting one will
    be.

    Parameters
    ----------
    vector : array of shape (,3)
        polarization vector cartesian coordinates in laser frame
        (see documentation for its exact definition)
    """

    def __init__(self, vector: np.ndarray):
        super(Vector, self).__init__()
        self.type = "Vector"
        self.vector = vector

    @property
    def vector(self):
        """array of shape (3,) : polarization vector cartesian coordinates in laser frame"""
        vector = np.asanyarray(self._vector)
        return vector

    @vector.setter
    def vector(self, value: np.ndarray) -> None:
        # convert to array
        value = np.asanyarray(value)
        if value.size != 3:
            raise ValueError("'vector' should be of size 3")
        # normalize
        norm = np.linalg.norm(value)
        if norm == 0:
            raise ValueError("Wrong value for 'vector'': norm is zero")
        self._vector = value / norm
        self.refresh_polarization_vector_angles()



# -*- coding: utf-8 -*-
"""Useful functions for plotting
"""

# %% IMPORTS
import numpy as np
import matplotlib.pyplot as plt

# %% COLLECTION FOR 3D ARROW PLOTTING
# source : https://gist.github.com/WetHat/1d6cd0f7309535311a539b42cccca89c

from matplotlib.patches import FancyArrowPatch
from matplotlib.text import Annotation
from mpl_toolkits.mplot3d.proj3d import proj_transform
from mpl_toolkits.mplot3d.axes3d import Axes3D


class Arrow3D(FancyArrowPatch):

    def __init__(self, x, y, z, dx, dy, dz, *args, **kwargs):
        super().__init__((0, 0), (0, 0), *args, **kwargs)
        self._xyz = (x, y, z)
        self._dxdydz = (dx, dy, dz)

    def draw(self, renderer):
        x1, y1, z1 = self._xyz
        dx, dy, dz = self._dxdydz
        x2, y2, z2 = (x1 + dx, y1 + dy, z1 + dz)

        xs, ys, zs = proj_transform((x1, x2), (y1, y2), (z1, z2), self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        super().draw(renderer)

    def do_3d_projection(self, renderer=None):
        x1, y1, z1 = self._xyz
        dx, dy, dz = self._dxdydz
        x2, y2, z2 = (x1 + dx, y1 + dy, z1 + dz)

        xs, ys, zs = proj_transform((x1, x2), (y1, y2), (z1, z2), self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))

        return np.min(zs)


class Annotation3D(Annotation):

    def __init__(self, text, xyz, *args, **kwargs):
        super().__init__(text, xy=(0, 0), *args, **kwargs)
        self._xyz = xyz

    def draw(self, renderer):
        x2, y2, z2 = proj_transform(*self._xyz, self.axes.M)
        self.xy = (x2, y2)
        super().draw(renderer)


def _arrow3D(ax, x, y, z, dx, dy, dz, *args, **kwargs):
    """Add an 3d arrow to an `Axes3D` instance."""

    arrow = Arrow3D(x, y, z, dx, dy, dz, *args, **kwargs)
    ax.add_artist(arrow)


def _annotate3D(ax, text, xyz, *args, **kwargs):
    """Add anotation `text` to an `Axes3d` instance."""

    annotation = Annotation3D(text, xyz, *args, **kwargs)
    ax.add_artist(annotation)


setattr(Axes3D, "arrow3D", _arrow3D)
setattr(Axes3D, "annotate3D", _annotate3D)



"""The ``atomsmltr.environment.fields.force`` subpackage provides definitions for forces

Those are mostly direct implementations of generic `Fields` objects
from `atomsmltr.environment.fields.generic`

Examples
---------

Setup a gravitational force

.. code-block:: python

    import numpy as np
    from atomsmltr.environment import ConstantForce
    from atomsmltr.atoms import Ytterbium

    m = Ytterbium().mass  # kg
    g = 9.81  # m/s^2
    direction = np.array([0, 0, -1])  # along -z
    grav_force = m * g * direction

    gravity = ConstantForce(field_value=grav_force, tag="gravity")

See also
--------
atomsmltr.environment.fields.generic

"""

__all__ = [
    "GradientForce",
    "ConstantForce",
    "Force",
]

from .force import (
    GradientForce,
    ConstantForce,
    Force,
)



"""Plottable objects
======================


"""

# % IMPORTS
from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
import numpy as np

from matplotlib.axes import Axes

# % LOCAL IMPORTS
from .tools import Axes3D


# % ABSTRACT CLASS


class Plottable(ABC):
    """This class defines several methods that should be common to
    all objects that contain data to be plotted"""

    def __init__(self):
        super(Plottable, self).__init__()

    def _init_ax(self, ax=None, ax3D=False):
        if ax is None:
            fig = plt.figure()
            if ax3D:
                ax = fig.add_subplot(111, projection="3d")
            else:
                ax = fig.add_subplot(111)
        return ax

    @abstractmethod
    def plot1D(self, ax=None):
        pass

    @abstractmethod
    def plot2D(self, ax=None, plane="XY"):
        pass

    @abstractmethod
    def plot3D(self, ax=None):
        pass

    def _process_2D_plot_args(self, ax, plane, limits, Npoints, cut):
        # ------------------------- START ARGUMENT CHECKING ----------------
        # - check plot config
        IMPLEMENTED_PLANES = ["XY", "YZ", "ZX"]
        if plane.upper() not in IMPLEMENTED_PLANES:
            raise ValueError(f"`plane` argument should be in {IMPLEMENTED_PLANES}")

        assert ax is None or isinstance(ax, Axes), "'ax' should be a matplotlib axis."
        # - check grid config
        # limits
        assert np.asanyarray(limits).size == 4, "`limits` should be an array of size 4"
        # Npoints
        Npoints = np.asanyarray(Npoints)
        msg = "`Npoints` should be an int or a list of three ints"
        assert Npoints.size in [1, 2], msg
        assert issubclass(Npoints.dtype.type, np.integer), msg
        # cut
        assert np.isscalar(cut), "'cut' should be a scalar"

        # ------------------------- STOP ARGUMENT CHECKING ----------------
        # - init ax (if needed)
        ax = self._init_ax(ax)

        # - init meshgrid
        xmin, xmax, ymin, ymax = limits
        Nx, Ny = (Npoints, Npoints) if Npoints.size == 1 else Npoints
        # depending on plane
        match plane.upper():
            case "XY":
                grid = np.mgrid[
                    xmin : xmax : Nx * 1j, ymin : ymax : Ny * 1j, cut:cut:1j
                ]
                position = grid.T[0]
                X, Y, _ = position.T
            case "YZ":
                grid = np.mgrid[
                    cut:cut:1j, xmin : xmax : Nx * 1j, ymin : ymax : Ny * 1j
                ]
                position = grid.T
                position = np.moveaxis(position, 2, 0)
                position = position[0]
                _, X, Y = position.T

            case "ZX":
                grid = np.mgrid[
                    ymin : ymax : Ny * 1j, cut:cut:1j, xmin : xmax : Nx * 1j
                ]
                position = grid.T
                position = np.moveaxis(position, 1, 0)
                position = position[0]
                Y, _, X = position.T

        return ax, position, X.T, Y.T
